"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.core.transform.terraform.converter.tf_resource_converter import \
    TerraformResourceConverter
from syndicate.core.transform.terraform.tf_resource_reference_builder import \
    build_cognito_identity_pool_id, build_role_arn_ref


class CognitoConverter(TerraformResourceConverter):

    def convert(self, name, resource):
        provider_name = resource.get('provider_name')
        open_id_provider_names = resource.get('open_id_providers', [])
        open_id_arns = ['arn:aws:iam::{0}:oidc-provider/{1}'.format(
            self.config.account_id, n) for n in open_id_provider_names]
        identity_pool = cognito_identity_pool(identity_pool_name=name,
                                              provider_name=provider_name,
                                              open_id_connect_provider_arns=open_id_arns)
        self.template.add_aws_cognito_identity_pool(meta=identity_pool)

        auth_role = resource.get('auth_role')
        unauth_role = resource.get('unauth_role')
        pool_attachment = aws_cognito_identity_pool_attachment(
            identity_pool_name=name,
            authenticated=auth_role,
            unauthenticated=unauth_role)
        self.template.add_aws_cognito_identity_pool_roles_attachment(
            pool_attachment)


def cognito_identity_pool(identity_pool_name,
                          allow_unauthenticated_identities=False,
                          provider_name=None,
                          login_providers=None,
                          open_id_connect_provider_arns=None,
                          cognito_identity_providers=None,
                          saml_provider_arns=None):
    identity_pool = {
        "identity_pool_name": identity_pool_name
    }

    if allow_unauthenticated_identities:
        identity_pool[
            'allow_unauthenticated_identities'] = allow_unauthenticated_identities
    if provider_name:
        identity_pool['developer_provider_name'] = provider_name
    if login_providers:
        identity_pool['supported_login_providers'] = login_providers
    if open_id_connect_provider_arns:
        identity_pool[
            'openid_connect_provider_arns'] = open_id_connect_provider_arns
    if cognito_identity_providers:
        identity_pool[
            'cognito_identity_providers'] = cognito_identity_providers
    if saml_provider_arns:
        identity_pool['saml_provider_arns'] = saml_provider_arns

    resource = {
        identity_pool_name: identity_pool
    }
    return resource


def aws_cognito_identity_pool_attachment(identity_pool_name, authenticated,
                                         unauthenticated):
    identity_pool_id_ref = build_cognito_identity_pool_id(
        pool_name=identity_pool_name)
    attachment = {
        "identity_pool_id": identity_pool_id_ref
    }

    roles = {}
    if authenticated:
        auth_role_ref = build_role_arn_ref(role_name=authenticated)
        roles['authenticated'] = auth_role_ref

    if unauthenticated:
        unauth_role_ref = build_role_arn_ref(role_name=unauthenticated)
        roles['unauthenticated']: unauth_role_ref

    if roles:
        attachment.update({'roles': roles})

    resource = {
        "attachment": attachment
    }
    return resource



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
IAM_POLICY = 'iam_policy'
IAM_ROLE = 'iam_role'
LAMBDA_TYPE = 'lambda'
LAMBDA_LAYER_TYPE = 'lambda_layer'
DYNAMO_TABLE_TYPE = 'dynamodb_table'
DAX_CLUSTER_TYPE = 'dax_cluster'
S3_BUCKET_TYPE = 's3_bucket'
CLOUD_WATCH_RULE_TYPE = 'cloudwatch_rule'
EVENT_BRIDGE_RULE_TYPE = 'eventbridge_rule'
EVENT_BRIDGE_SCHEDULE_TYPE = 'eventbridge_schedule'
SQS_QUEUE_TYPE = 'sqs_queue'
API_GATEWAY_TYPE = 'api_gateway'
API_GATEWAY_OAS_V3_TYPE = 'api_gateway_oas_v3'
WEB_SOCKET_API_GATEWAY_TYPE = 'web_socket_api_gateway'
SWAGGER_UI_TYPE = 'swagger_ui'
COGNITO_FEDERATED_POOL_TYPE = 'cognito_federated_pool'
COGNITO_USER_POOL_TYPE = 'cognito_idp'
SNS_TOPIC_TYPE = 'sns_topic'
SNS_PLATFORM_APPLICATION_TYPE = 'sns_application'
CLOUD_WATCH_ALARM_TYPE = 'cloudwatch_alarm'
EBS_TYPE = 'beanstalk_app'
STEP_FUNCTION_TYPE = 'step_functions'
KINESIS_STREAM_TYPE = 'kinesis_stream'
STATE_ACTIVITY_TYPE = 'state_activity'
EC2_INSTANCE_TYPE = 'ec2_instance'
EC2_LAUNCH_TEMPLATE_TYPE = 'ec2_launch_template'
BATCH_COMPENV_TYPE = 'batch_compenv'
BATCH_JOBQUEUE_TYPE = 'batch_jobqueue'
BATCH_JOBDEF_TYPE = 'batch_jobdef'
FIREHOSE_TYPE = 'firehose'
DOCUMENTDB_CLUSTER_TYPE = 'documentdb_cluster'
DOCUMENTDB_INSTANCE_TYPE = 'documentdb_instance'

S3_PATH_NAME = 's3_path'
EXPORT_DIR_NAME = 'export'

# == BUILD PARAMS =============================================================
ARTIFACTS_FOLDER = 'bundles'
BUILD_META_FILE_NAME = 'build_meta.json'
LAMBDA_CONFIG_FILE_NAME = 'lambda_config.json'
LAMBDA_LAYER_CONFIG_FILE_NAME = 'lambda_layer_config.json'
REQ_FILE_NAME = 'requirements.txt'
NODE_REQ_FILE_NAME = 'package.json'
LOCAL_REQ_FILE_NAME = 'local_requirements.txt'
RESOURCES_FILE_NAME = 'deployment_resources.json'
OAS_V3_FILE_NAME = 'oas_v3.json'
SWAGGER_UI_SPEC_NAME_TEMPLATE = '{name}_spec.json'
SWAGGER_UI_ARTIFACT_NAME_TEMPLATE = 'swagger_ui_{name}.zip'
SWAGGER_UI_CONFIG_FILE_NAME = 'swagger_ui_config.json'
# layer.zip
# │ python/PIL
# └ python/Pillow-5.3.0.dist-info
PYTHON_LAMBDA_LAYER_PATH = 'python'
NODE_LAMBDA_LAYER_PATH = 'nodejs'

DEFAULT_SEP = '/'

DEPLOY_RESOURCE_TYPE_PRIORITY = {
    IAM_POLICY: 1,
    IAM_ROLE: 2,
    DYNAMO_TABLE_TYPE: 3,
    DAX_CLUSTER_TYPE: 4,
    S3_BUCKET_TYPE: 5,
    CLOUD_WATCH_RULE_TYPE: 6,
    EVENT_BRIDGE_RULE_TYPE: 6,
    SNS_TOPIC_TYPE: 7,
    SQS_QUEUE_TYPE: 8,
    KINESIS_STREAM_TYPE: 9,
    CLOUD_WATCH_ALARM_TYPE: 10,
    LAMBDA_LAYER_TYPE: 11,
    LAMBDA_TYPE: 12,
    STATE_ACTIVITY_TYPE: 13,
    STEP_FUNCTION_TYPE: 14,
    COGNITO_USER_POOL_TYPE: 15,
    API_GATEWAY_TYPE: 16,
    API_GATEWAY_OAS_V3_TYPE: 17,
    WEB_SOCKET_API_GATEWAY_TYPE: 18,
    SWAGGER_UI_TYPE: 19,
    COGNITO_FEDERATED_POOL_TYPE: 19,
    EC2_LAUNCH_TEMPLATE_TYPE: 19,
    EBS_TYPE: 20,
    EC2_INSTANCE_TYPE: 21,
    SNS_PLATFORM_APPLICATION_TYPE: 22,
    BATCH_COMPENV_TYPE: 23,
    BATCH_JOBQUEUE_TYPE: 24,
    BATCH_JOBDEF_TYPE: 25,
    FIREHOSE_TYPE: 26,
    DOCUMENTDB_CLUSTER_TYPE: 27,
    DOCUMENTDB_INSTANCE_TYPE: 28,
    EVENT_BRIDGE_SCHEDULE_TYPE: 29
}

CLEAN_RESOURCE_TYPE_PRIORITY = {
    IAM_ROLE: 1,
    IAM_POLICY: 2,
    DAX_CLUSTER_TYPE: 3,
    DYNAMO_TABLE_TYPE: 4,
    SWAGGER_UI_TYPE: 4,
    S3_BUCKET_TYPE: 5,
    CLOUD_WATCH_RULE_TYPE: 6,
    EVENT_BRIDGE_RULE_TYPE: 6,
    SNS_TOPIC_TYPE: 7,
    SQS_QUEUE_TYPE: 8,
    KINESIS_STREAM_TYPE: 9,
    CLOUD_WATCH_ALARM_TYPE: 10,
    LAMBDA_TYPE: 11,
    LAMBDA_LAYER_TYPE: 12,
    STATE_ACTIVITY_TYPE: 13,
    STEP_FUNCTION_TYPE: 14,
    COGNITO_USER_POOL_TYPE: 15,
    WEB_SOCKET_API_GATEWAY_TYPE: 16,
    API_GATEWAY_TYPE: 17,
    API_GATEWAY_OAS_V3_TYPE: 18,
    COGNITO_FEDERATED_POOL_TYPE: 19,
    EBS_TYPE: 20,
    EC2_INSTANCE_TYPE: 21,
    SNS_PLATFORM_APPLICATION_TYPE: 22,
    BATCH_JOBDEF_TYPE: 23,
    BATCH_JOBQUEUE_TYPE: 24,
    BATCH_COMPENV_TYPE: 25,
    EC2_LAUNCH_TEMPLATE_TYPE: 25,
    FIREHOSE_TYPE: 26,
    DOCUMENTDB_INSTANCE_TYPE: 27,
    DOCUMENTDB_CLUSTER_TYPE: 28,
    EVENT_BRIDGE_SCHEDULE_TYPE: 29
}

UPDATE_RESOURCE_TYPE_PRIORITY = {
    IAM_POLICY: 1,
    IAM_ROLE: 2,
    DYNAMO_TABLE_TYPE: 3,
    LAMBDA_LAYER_TYPE: 4,
    LAMBDA_TYPE: 5,
    API_GATEWAY_OAS_V3_TYPE: 6,
    EC2_LAUNCH_TEMPLATE_TYPE: 6,
    BATCH_JOBDEF_TYPE: 7,
    BATCH_COMPENV_TYPE: 8,
    SWAGGER_UI_TYPE: 9
}

RESOURCE_LIST = list(DEPLOY_RESOURCE_TYPE_PRIORITY.keys())
DATE_FORMAT_ISO_8601 = '%Y-%m-%dT%H:%M:%SZ'

TEST_ACTION = 'test'
BUILD_ACTION = 'build'
DEPLOY_ACTION = 'deploy'
UPDATE_ACTION = 'update'
CLEAN_ACTION = 'clean'
PARTIAL_CLEAN_ACTION = 'partial_clean'
SYNC_ACTION = 'sync'
STATUS_ACTION = 'status'
WARMUP_ACTION = 'warmup'
PROFILER_ACTION = 'profiler'
ASSEMBLE_JAVA_MVN_ACTION = 'assemble_java_mvn'
ASSEMBLE_PYTHON_ACTION = 'assemble_python'
ASSEMBLE_NODE_ACTION = 'assemble_node'
ASSEMBLE_SWAGGER_UI_ACTION = 'assemble_swagger_ui'
ASSEMBLE_ACTION = 'assemble'
PACKAGE_META_ACTION = 'package_meta'
CREATE_DEPLOY_TARGET_BUCKET_ACTION = 'create_deploy_target_bucket'
UPLOAD_ACTION = 'upload'
COPY_BUNDLE_ACTION = 'copy_bundle'
EXPORT_ACTION = 'export'

NONE_AUTH_TYPE, IAM_AUTH_TYPE = 'NONE', 'AWS_IAM'

MANY_LINUX_2014_PLATFORM = 'manylinux2014_x86_64'
OPTIMAL_INSTANCE_TYPE = 'optimal'

POSSIBLE_RETENTION_DAYS = (
1, 3, 5, 7, 14, 30, 60, 90, 120, 150, 180, 365, 400, 545, 731, 1096, 1827, 2192,
2557, 2922, 3288, 3653)
DEFAULT_LOGS_EXPIRATION = 30
SOURCE_ARN_DEEP_KEY = ('Condition', 'ArnLike', 'AWS:SourceArn')
SECURITY_SCHEMAS_DEEP_KEY = ('components', 'securitySchemes')
LAMBDA_ARCHITECTURE_LIST = ['x86_64', 'arm64']

API_GW_DEFAULT_THROTTLING_RATE_LIMIT = 10000
API_GW_DEFAULT_THROTTLING_BURST_LIMIT = 5000
COGNITO_USER_POOL_AUTHORIZER_TYPE = 'COGNITO_USER_POOLS'
REQUEST_LAMBDA_AUTHORIZER_TYPE = 'REQUEST'
TOKEN_LAMBDA_AUTHORIZER_TYPE = 'TOKEN'
API_GW_AUTHORIZER_TYPES = [COGNITO_USER_POOL_AUTHORIZER_TYPE,
                           TOKEN_LAMBDA_AUTHORIZER_TYPE,
                           REQUEST_LAMBDA_AUTHORIZER_TYPE]
CUSTOM_AUTHORIZER_KEY = 'CUSTOM'

S3_BUCKET_ACL_LIST = ['private', 'public-read',
                      'public-read-write', 'authenticated-read']


SYNDICATE_WIKI_PAGE = 'https://github.com/epam/aws-syndicate/wiki/'
SYNDICATE_PROJECT_EXAMPLES_PAGE = 'https://github.com/epam/aws-syndicate/tree/master/examples/'

JAVA_LAMBDAS_WIKI_PAGE = '3.-Lambda-Requirements-for-Automatic-Articfacts-Build#32-java-lambdas'

EC2_LAUNCH_TEMPLATE_SUPPORTED_IMDS_VERSIONS = ['v1.0', 'v2.0']



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import getpass
import os
import re
import sys
import time
from datetime import datetime, timedelta
from pathlib import Path, PurePath
from typing import Union

import yaml

from syndicate.commons.log_helper import get_logger
from syndicate.core.constants import BUILD_ACTION, \
    DEPLOY_ACTION, UPDATE_ACTION, CLEAN_ACTION, PACKAGE_META_ACTION, \
    PARTIAL_CLEAN_ACTION
from syndicate.core.constants import DATE_FORMAT_ISO_8601
from syndicate.core.groups import RUNTIME_JAVA, RUNTIME_NODEJS, RUNTIME_PYTHON, \
    RUNTIME_SWAGGER_UI

CAPITAL_LETTER_REGEX = '[A-Z][^A-Z]*'

STATE_NAME = 'name'
STATE_LOCKS = 'locks'
STATE_LAMBDAS = 'lambdas'
STATE_BUILD_PROJECT_MAPPING = 'build_projects_mapping'
STATE_LOG_EVENTS = 'events'
STATE_LATEST_DEPLOY = 'latest_deploy'
LOCK_LOCKED_TILL = 'locked_till'
LOCK_LAST_MODIFICATION_DATE = 'last_modification_date'
LOCK_INITIATOR = 'initiator'

MODIFICATION_LOCK = 'modification_lock'
WARMUP_LOCK = 'warm_up_lock'
PROJECT_STATE_FILE = '.syndicate'
INIT_FILE = '__init__.py'
INDEX_FILE = 'index.js'

BUILD_MAPPINGS = {
    RUNTIME_JAVA: 'jsrc/main/java',
    RUNTIME_PYTHON: 'src',
    RUNTIME_NODEJS: 'app',
    RUNTIME_SWAGGER_UI: 'swagger_src'
}

OPERATION_LOCK_MAPPINGS = {
    'deploy': MODIFICATION_LOCK
}
KEEP_EVENTS_DAYS = 30
LEAVE_LATEST_EVENTS = 20

_LOG = get_logger('project-state')


class ProjectState:

    def __init__(self, project_path: str = None, dct: dict = None):
        """In case 'dct' param is given it will be assigned to the internal
        _dict variable instead of loading from a file. It comes handy when
        we need to get ProjectState object without loading from file. All
        the existing functionality remains unimpaired"""
        if not (project_path or dct):
            message = 'Either project_path or dct of both must be ' \
                      'specified!'
            _LOG.error(message)
            raise AssertionError(message)
        if project_path:
            self.project_path = project_path
            self.state_path = os.path.join(project_path, PROJECT_STATE_FILE)
        self.dct = dct if dct else self.__load_project_state_file()

    @staticmethod
    def generate(project_path, project_name):
        project_state = dict(name=project_name)
        with open(os.path.join(project_path, PROJECT_STATE_FILE),
                  'w') as state_file:
            yaml.dump(project_state, state_file)
        return ProjectState(project_path=project_path)

    @staticmethod
    def check_if_project_state_exists(project_path):
        return os.path.exists(os.path.join(project_path, PROJECT_STATE_FILE))

    @property
    def dct(self) -> dict:
        return self._dict

    @dct.setter
    def dct(self, dct: dict):
        self._dict = dct

    @staticmethod
    def get_remote() -> 'ProjectState':
        from syndicate.core import CONN, CONFIG
        bucket_name = CONFIG.deploy_target_bucket
        key_compound = PurePath(
            CONFIG.deploy_target_bucket_key_compound, PROJECT_STATE_FILE
        ).as_posix()
        s3 = CONN.s3()
        remote_project_state = s3.load_file_body(bucket_name=bucket_name,
                                                 key=key_compound)
        remote_project_state = yaml.unsafe_load(remote_project_state)
        _LOG.info(f'Unsafely loaded project state file from S3 bucket. The '
                  f'retrieved object has type: '
                  f'{type(remote_project_state).__name__}')
        if isinstance(remote_project_state, dict):
            remote_project_state = ProjectState(dct=remote_project_state)
            _LOG.info(f'Made ProjectState object from the dict loaded from S3 '
                      f'bucket')
        else:  # isinstance(remote_project_state, ProjectState):
            _LOG.warning(f'Loaded project state object is already instance of '
                         f'ProjectState. Likely .syndicate from the '
                         f'the bucket is obsolete. Rewriting...')
        return remote_project_state

    def save_to_remote(self, project_state_to_save: 'ProjectState' = None):
        dict_to_save = project_state_to_save.dct if \
            project_state_to_save else self.dct
        from syndicate.core import CONN, CONFIG
        bucket_name = CONFIG.deploy_target_bucket
        key_compound = PurePath(CONFIG.deploy_target_bucket_key_compound,
                                PROJECT_STATE_FILE).as_posix()
        s3 = CONN.s3()
        s3.put_object(file_obj=yaml.dump(dict_to_save, sort_keys=False),
                      key=key_compound,
                      bucket=bucket_name,
                      content_type='application/x-yaml')

    def save(self):
        with open(self.state_path, 'w') as state_file:
            yaml.dump(self.dct, state_file, sort_keys=False)

    @property
    def name(self):
        return self.dct.get(STATE_NAME)

    @property
    def default_deploy_name(self):
        import re
        parts = []
        if '_' in self.name:
            parts.extend(self.name.split('_'))
        if not parts:
            parts = re.findall(CAPITAL_LETTER_REGEX, self.name)
        if not parts:
            parts = [self.name]
        return '-'.join([_.lower() for _ in parts])

    @name.setter
    def name(self, name):
        self.dct.update({STATE_NAME: name})

    @property
    def locks(self):
        locks = self.dct.get(STATE_LOCKS)
        if not locks:
            locks = dict()
            self.dct.update({STATE_LOCKS: locks})
        return locks

    @property
    def lambdas(self):
        lambdas = self.dct.get(STATE_LAMBDAS)
        if not lambdas:
            return dict()
        return lambdas

    @property
    def events(self):
        events = self.dct.get(STATE_LOG_EVENTS)
        if not events:
            events = []
            self.dct.update({STATE_LOG_EVENTS: events})
        return events

    @events.setter
    def events(self, events):
        self.dct.update({STATE_LOG_EVENTS: events})

    @property
    def latest_deploy(self):
        latest_deploy = self.dct.get(STATE_LATEST_DEPLOY)
        if not latest_deploy:
            latest_deploy = {}
            self.dct.update({STATE_LATEST_DEPLOY: latest_deploy})
        return latest_deploy

    @latest_deploy.setter
    def latest_deploy(self, latest_deploy):
        self.dct[STATE_LATEST_DEPLOY] = latest_deploy

    @property
    def latest_bundle_name(self):
        """Returns bundle_name from the one of the latest operations which
        can guarantee that the bundle is ready"""
        operations = [BUILD_ACTION, PACKAGE_META_ACTION]
        for event in self.events:
            if event.get('operation') in operations:
                bundle_name = event.get('bundle_name')
                if bundle_name:
                    return bundle_name

    @property
    def latest_deployed_bundle_name(self):
        return self.latest_deploy.get('bundle_name')

    @property
    def latest_deployed_deploy_name(self):
        return self.latest_deploy.get('deploy_name')

    def _get_attribute_from_latest_operation(self, operation_name, attribute):
        events = self.events
        event = next((event for event in events if
                      event.get('operation') == operation_name), None)
        if event:
            return event.get(attribute)

    @property
    def latest_modification(self):
        events = self.events
        modification_ops = [DEPLOY_ACTION, UPDATE_ACTION, CLEAN_ACTION]
        latest = next((event for event in events if
                       event.get('operation') in modification_ops), None)
        return latest

    def is_lock_free(self, lock_name):
        lock = self.locks.get(lock_name)
        if not lock:
            return True
        elif not lock.get(LOCK_LOCKED_TILL):
            return True
        elif locked_till := lock.get(LOCK_LOCKED_TILL):
            locked_till_datetime = datetime.strptime(
                locked_till, DATE_FORMAT_ISO_8601)
            if datetime.timestamp(locked_till_datetime) <= time.time():
                lock[LOCK_LOCKED_TILL] = None
                return True
        return False

    def acquire_lock(self, lock_name):
        self.__modify_lock_state(lock_name, True)

    def release_lock(self, lock_name):
        self.__modify_lock_state(lock_name, False)

    def actualize_locks(self, other_project_state):
        locks = self.locks
        other_locks = other_project_state.locks
        all_lock_names = set(locks.keys()).union(set(other_locks.keys()))
        for lock_name in all_lock_names:
            lock = locks.get(lock_name)
            other_lock = other_locks.get(lock_name)
            if lock is None:
                locks.update({lock_name: other_lock})
            elif other_lock is None:
                other_locks.update({lock_name: lock})
            elif (lock.get(LOCK_LAST_MODIFICATION_DATE) <
                  other_lock.get(LOCK_LAST_MODIFICATION_DATE)):
                locks.update({lock_name: other_lock})
            else:
                other_locks.update({lock_name: lock})

    def actualize_latest_deploy(self, other_project_state: 'ProjectState'):
        local_deploy = self.latest_deploy
        remote_deploy = other_project_state.latest_deploy
        if local_deploy and remote_deploy:
            local_time_start = datetime.strptime(local_deploy['time_start'],
                                                 DATE_FORMAT_ISO_8601)
            remote_time_start = datetime.strptime(remote_deploy['time_start'],
                                                  DATE_FORMAT_ISO_8601)
            if remote_time_start > local_time_start:
                self.latest_deploy = remote_deploy
        elif remote_deploy:
            self.latest_deploy = remote_deploy

    def refresh_lambda_state(self):
        """
        Refreshes current Project lambda State, be resolving
        the compatibility with the retained state. Given any consistency
        conflict the ProjectState is re-persisted.
        :return: None
        """
        from syndicate.core.generators.lambda_function import \
            resolve_lambda_path
        from syndicate.core.helper import check_lambda_state_consistency

        _persistence_need: bool = False
        _project_path = Path(self.project_path)
        _stale_lambdas = self.lambdas
        for runtime, source in BUILD_MAPPINGS.items():
            _path = resolve_lambda_path(_project_path, runtime, source)
            _updated = self._update_lambdas_from_path(_path, runtime)
            if not _persistence_need and check_lambda_state_consistency(
                    objected_lambdas=_updated,
                    subjected_lambdas=_stale_lambdas,
                    runtime=runtime
            ):
                _persistence_need = True

        if _persistence_need:
            self.save()

    def _update_lambdas_from_path(self, path: Union[str, Path], runtime: str):
        """
        Non persistently updates ProjectState runtime and
        any found lambdas from a given path.
        :parameter path:Path
        :parameter runtime: str
        :return: List
        """
        try:
            path = path if isinstance(path, Path) else Path(path)
        except (TypeError, Exception):
            _LOG.error(f'Requested path {path} must be of str or Path type.')
            return []

        _LOG.info(f'Going to resolve any lambda names from a given path: '
                  f'{path.absolute()}.')
        _lambdas: list = self._resolve_lambdas_from_path(path, runtime)
        for name in self._resolve_lambdas_from_path(path, runtime):
            _LOG.info(f'Going to add the following \'{runtime}\' lambda:'
                      f'\'{name}\' to the pending ProjectState.')
            self.add_lambda(lambda_name=name, runtime=runtime)
        if _lambdas:
            self.add_project_build_mapping(runtime)
        return _lambdas

    def add_lambda(self, lambda_name, runtime):
        lambdas = self.dct.get(STATE_LAMBDAS)
        if not lambdas:
            lambdas = dict()
            self.dct.update({STATE_LAMBDAS: lambdas})
        lambdas.update({lambda_name: {'runtime': runtime}})

    def add_project_build_mapping(self, runtime):
        build_project_mappings = self.dct.get(STATE_BUILD_PROJECT_MAPPING)
        if not build_project_mappings:
            build_project_mappings = dict()
            self.dct.update(
                {STATE_BUILD_PROJECT_MAPPING: build_project_mappings})
        build_mapping = BUILD_MAPPINGS.get(runtime)
        build_project_mappings.update({runtime: build_mapping})

    def load_project_build_mapping(self):
        return self.dct.get(STATE_BUILD_PROJECT_MAPPING)

    def log_execution_event(self, **kwargs):
        operation = kwargs.get('operation')

        operation_status = kwargs.get('operation_status')
        rollback_on_error = kwargs.get('rollback_on_error')
        if not isinstance(operation_status, bool):
            kwargs.pop('operation_status')

        if operation in [DEPLOY_ACTION, PARTIAL_CLEAN_ACTION]:
            params = kwargs.copy()
            params.pop('operation')

            if not(operation_status is False and rollback_on_error is True):
                params.pop('rollback_on_error')
                self._set_latest_deploy_info(**params)

        if operation == CLEAN_ACTION:
            self._delete_latest_deploy_info()

        kwargs = {
            key: value for key, value in kwargs.items() if value is not None
        }
        self.events.insert(0, kwargs)
        self.__save_events()

    def _set_latest_deploy_info(self, **kwargs):
        kwargs = {
            key: value for key, value in kwargs.items() if value is not None
        }
        self.latest_deploy = kwargs

    def _delete_latest_deploy_info(self):
        self.latest_deploy = {}
        remote_project_state = ProjectState.get_remote()
        remote_project_state.latest_deploy = {}
        self.save_to_remote(project_state_to_save=remote_project_state)

    def add_execution_events(self, events):
        for event in events:
            if event not in self.events:
                self.events.append(event)
        self.events.sort(key=lambda e: e.get('time_end'), reverse=True)
        self.__save_events()

    def __modify_lock_state(self, lock_name, locked):
        from syndicate.core import CONFIG
        locked_till = CONFIG.lock_lifetime_minutes

        locks = self.locks
        lock = locks.get(lock_name)
        modification_datetime = datetime.fromtimestamp(time.time())
        timestamp = modification_datetime.strftime(DATE_FORMAT_ISO_8601)
        locked_till_timestamp = (modification_datetime +
                                 timedelta(minutes=locked_till)).strftime(
            DATE_FORMAT_ISO_8601)

        modified_lock = {
            LOCK_LAST_MODIFICATION_DATE: timestamp,
            LOCK_LOCKED_TILL: locked_till_timestamp if locked else None,
            LOCK_INITIATOR: getpass.getuser()
        }
        if lock:
            lock.update(modified_lock)
        else:
            locks.update({lock_name: modified_lock})
        self.save()

    def __load_project_state_file(self):
        if not ProjectState.check_if_project_state_exists(self.project_path):
            raise AssertionError(
                f'There is no .syndicate file in {self.project_path}')
        with open(self.state_path) as state_file:
            return yaml.safe_load(state_file.read())

    def __save_events(self):
        current_time = datetime.fromtimestamp(time.time())
        index_out_days = None
        for i, event in enumerate(self.events):
            if (current_time -
                datetime.strptime(event.get('time_end'),
                                  DATE_FORMAT_ISO_8601)).days > KEEP_EVENTS_DAYS:
                index_out_days = i
                break

        if index_out_days:
            if index_out_days >= LEAVE_LATEST_EVENTS:
                self.events = self.events[:index_out_days]
            else:
                self.events = self.events[:LEAVE_LATEST_EVENTS]
        self.save()

    @staticmethod
    def build_from_structure(config):
        """Builds project state file from existing project folder in case of
        moving from older versions
        :type config: syndicate.core.conf.processor.ConfigHolder
        """
        from syndicate.core.generators.lambda_function import \
            resolve_lambda_path
        absolute_path = config.project_path
        project_path = Path(absolute_path)
        project_state = ProjectState.generate(project_path=absolute_path,
                                              project_name=project_path.name)

        for runtime, source_path in BUILD_MAPPINGS.items():
            lambdas_path = resolve_lambda_path(project_path, runtime,
                                               source_path)
            if os.path.exists(lambdas_path):
                project_state.add_project_build_mapping(runtime)
                project_state._update_lambdas_from_path(lambdas_path, runtime)

        project_state.save()

        return project_state

    @staticmethod
    def _resolve_lambdas_from_path(path: Path, runtime: str):
        """
        Resolves a list of names bound to lambda functions,
        retained inside a given path, based on a provided runtime.
        :parameter path: Path
        :parameter runtime: str
        :return: List[str]
        """

        lambda_list = []
        _java_lambda_regex = 'lambdaName\s*=\s*"(.+)"'

        if not path.exists():
            return lambda_list

        for item in path.iterdir():
            if runtime == RUNTIME_JAVA:
                if not item.is_file():
                    continue
                try:
                    match = re.search(_java_lambda_regex, item.read_text())
                    if match:
                        lambda_list.append(match.group(1))
                except (OSError, Exception):
                    print("Couldn't retrieve lambda name from the java "
                          "lambda by path: {}".format(item.absolute()),
                          file=sys.stderr)
            else:
                if (item/INIT_FILE).exists() or (item/INDEX_FILE).exists():
                    lambda_list.append(item.name)

        return lambda_list



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from json import dumps

from boto3 import client
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('syndicate.connection.step_functions_connection')


@apply_methods_decorator(retry())
class SFConnection(object):
    """ STS connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('stepfunctions', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new Step Functions connection.')

    def create_state_machine(self, machine_name, definition, role_arn):
        if isinstance(definition, dict):
            definition = dumps(definition)
        return self.client.create_state_machine(name=machine_name,
                                                definition=definition,
                                                roleArn=role_arn)

    def describe_state_machine(self, arn):
        try:
            return self.client.describe_state_machine(stateMachineArn=arn)
        except ClientError as e:
            if 'StateMachineDoesNotExist' in str(e):
                pass  # valid exception
            else:
                raise e

    def delete_state_machine(self, arn):
        return self.client.delete_state_machine(stateMachineArn=arn)

    def list_state_machines(self):
        result = []
        response = self.client.list_state_machines()
        do_continue = response.get('nextToken')
        if 'stateMachines' in response:
            result.extend([i['name'] for i in response['stateMachines']])
        while do_continue:
            response = self.client.list_state_machines(nextToken=do_continue)
            do_continue = response.get('nextToken')
            if 'stateMachines' in response:
                result.extend([i['name'] for i in response['stateMachines']])
        return result

    def list_executions_by_status(self, state_machine_arn, execution_status):
        result = []
        response = self.client.list_executions(
            stateMachineArn=state_machine_arn,
            statusFilter=execution_status
        )
        do_continue = response.get('nextToken')
        if 'executions' in response:
            result.extend(response['executions'])
        while do_continue:
            response = self.client.list_executions(
                stateMachineArn=state_machine_arn,
                statusFilter=execution_status,
                nextToken=do_continue
            )
            do_continue = response.get('nextToken')
            if 'executions' in response:
                result.extend(response['executions'])
        return result

    def stop_execution(self, execution_arn):
        return self.client.stop_execution(executionArn=execution_arn)

    def create_activity(self, name):
        return self.client.create_activity(name=name)

    def describe_activity(self, arn):
        try:
            return self.client.describe_activity(activityArn=arn)
        except ClientError as e:
            if 'ActivityDoesNotExist' in str(e):
                pass  # valid exception
            else:
                raise e

    def delete_activity(self, arn):
        self.client.delete_activity(activityArn=arn)

    def list_activities(self):
        result = []
        response = self.client.list_activities()
        do_continue = response.get('nextToken')
        if 'activities' in response:
            result.extend([i['name'] for i in response['activities']])
        while do_continue:
            response = self.client.list_activities(nextToken=do_continue)
            do_continue = response.get('nextToken')
            if 'activities' in response:
                result.extend([i['name'] for i in response['activities']])
        return result



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import concurrent
import copy
import functools
import json
from concurrent.futures import ALL_COMPLETED, ThreadPoolExecutor
from functools import cmp_to_key

from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.build.bundle_processor import (create_deploy_output,
                                                   load_deploy_output,
                                                   load_failed_deploy_output,
                                                   load_meta_resources,
                                                   remove_failed_deploy_output,
                                                   load_latest_deploy_output)
from syndicate.core.build.helper import _json_serial
from syndicate.core.build.meta_processor import (resolve_meta,
                                                 populate_s3_paths,
                                                 resolve_resource_name)
from syndicate.core.constants import (BUILD_META_FILE_NAME,
                                      CLEAN_RESOURCE_TYPE_PRIORITY,
                                      DEPLOY_RESOURCE_TYPE_PRIORITY,
                                      UPDATE_RESOURCE_TYPE_PRIORITY,
                                      PARTIAL_CLEAN_ACTION)
from syndicate.core.helper import exit_on_exception, prettify_json

BUILD_META = 'build_meta'
DEPLOYMENT_OUTPUT = 'deployment_output'

_LOG = get_logger('syndicate.core.build.deployment_processor')
USER_LOG = get_user_logger()


def get_dependencies(name, meta, resources_dict, resources):
    """ Get dependencies from resources that needed to create them too.

    :type name: str
    :type meta: dict
    :type resources_dict: dict
    :param resources:
    :param resources_dict: resources that will be created {name: meta}
    """
    resources_dict[name] = meta
    if meta.get('dependencies'):
        for dependency in meta.get('dependencies'):
            dep_name = dependency['resource_name']
            dep_meta = resources[dep_name]
            resources_dict[dep_name] = dep_meta
            if dep_meta.get('dependencies'):
                get_dependencies(dep_name, dep_meta, resources_dict, resources)


# todo implement resources sorter according to priority
# todo add dependency resolving
def _process_resources(resources, handlers_mapping, pass_context=False):
    output = {}
    args = []
    resource_type = None
    try:
        for res_name, res_meta in resources:
            current_res_type = res_meta['resource_type']

            if resource_type is None:
                resource_type = current_res_type

            if current_res_type == resource_type:
                args.append(_build_args(name=res_name,
                                        meta=res_meta,
                                        context=output,
                                        pass_context=pass_context))
                continue
            elif current_res_type != resource_type:
                USER_LOG.info(f'Processing {resource_type} resources')
                func = handlers_mapping[resource_type]
                response = func(args)
                process_response(response=response, output=output)

                del args[:]
                args.append(_build_args(name=res_name,
                                        meta=res_meta,
                                        context=output,
                                        pass_context=pass_context))
                resource_type = current_res_type
        if args:
            USER_LOG.info(f'Processing {resource_type} resources')
            func = handlers_mapping[resource_type]

            response = func(args)
            process_response(response=response, output=output)

        return True, output
    except Exception as e:
        USER_LOG.exception('Error occurred while {0} '
                           'resource creating: {1}'.format(resource_type,
                                                           str(e)))
        return False, output


def _build_args(name, meta, context, pass_context=False):
    """
    Builds parameters to pass to resource_type handler.
    Default parameters dict consists of name and meta keys.
    If pass_context set to True, parameters dict is extended with 'context' key
    :param name: name of the resource
    :param meta: definition of the resource
    :param context: result of previously deployed resources in scope of current syndicate execution
    :param pass_context: flag. Manages if output will be included to parameters
    :return: prepared parameters to be passed to resource_type handler.
    """
    params = {'name': name, 'meta': meta}
    if pass_context:
        params['context'] = context
    return params


def update_failed_output(res_name, res_meta, resource_type, output):
    from syndicate.core import PROCESSOR_FACADE

    try:
        describe_func = PROCESSOR_FACADE.describe_handlers()[resource_type]
        failed_resource_output = describe_func(res_name, res_meta)
        if failed_resource_output:
            if isinstance(failed_resource_output, list):
                for item in failed_resource_output:
                    output.update(item)
            else:
                output.update(failed_resource_output)
    except Exception as e:
        _LOG.warning(f'Unable to describe {resource_type} '
                     f'resource with name {res_name}. Exception: {e}')
    return output


def deploy_resources(resources):
    from syndicate.core import PROCESSOR_FACADE
    return _process_resources(
        resources=resources,
        handlers_mapping=PROCESSOR_FACADE.create_handlers())


def update_resources(resources):
    from syndicate.core import PROCESSOR_FACADE
    return _process_resources(
        resources=resources,
        handlers_mapping=PROCESSOR_FACADE.update_handlers(),
        pass_context=True)


def clean_resources(output):
    from syndicate.core import PROCESSOR_FACADE
    args = []
    resource_type = None
    # clean all resources
    for arn, config in output:
        res_type = config['resource_meta']['resource_type']
        if resource_type is None:
            resource_type = res_type

        if res_type == resource_type:
            args.append({'arn': arn, 'config': config})
            continue
        elif res_type != resource_type:
            USER_LOG.info('Removing {0} resources ...'.format(resource_type))
            func = PROCESSOR_FACADE.remove_handlers()[resource_type]
            func(args)
            del args[:]
            args.append({'arn': arn, 'config': config})
            resource_type = res_type
    if args:
        USER_LOG.info('Removing {0} resources ...'.format(resource_type))
        func = PROCESSOR_FACADE.remove_handlers()[resource_type]
        func(args)


# todo implement saving failed output
def continue_deploy_resources(resources, failed_output):
    from syndicate.core import PROCESSOR_FACADE
    handler_mappings = PROCESSOR_FACADE.create_handlers()

    resources_to_deploy = []
    deploy_result = True
    args = []

    failed_resource_names = {data['resource_name'] for data in
                             failed_output.values()}

    for resource_name, resource_meta in resources:
        if resource_name not in failed_resource_names:
            resources_to_deploy.append((resource_name, resource_meta))

    if not resources_to_deploy:
        return deploy_result, failed_output

    for resource in resources_to_deploy:
        res_name, res_meta = resource
        res_type = res_meta['resource_type']
        func = handler_mappings[res_type]
        try:
            if func:
                args.append(_build_args(
                    name=res_name,
                    meta=res_meta,
                    context={}
                ))
                USER_LOG.info(f'Processing {res_type} resources')
                response = func(args)
                del args[:]
                if response:
                    process_response(
                        response=response, output=failed_output
                    )
                else:
                    _LOG.warning(f"Handler for resource type: {res_type}"
                                 f" did not returned any response")
            else:
                _LOG.warning(f"Handler for the `{res_name}` resource of"
                             f" {res_type} type was not found. Resource"
                             f" has not been deployed.")
        except Exception as e:
            _LOG.exception(
                'Error occurred while {0} resource creating: {1}'.
                format(res_type, str(e)))
            deploy_result = False
            return deploy_result, failed_output


def process_response(response, output: dict):

    if isinstance(response, dict):
        output.update(response)
    elif isinstance(response, tuple):
        result, exceptions = response
        if result:
            output.update(result)
        raise Exception(exceptions[0])

    return output


def __move_output_content(args, failed_output, updated_output):
    for arg in args:
        resource_output = __find_output_by_resource_name(
            failed_output, arg['name'])
        if resource_output:
            updated_output.update(resource_output)


def __find_output_by_resource_name(output, resource_name):
    found_items = {}
    for k, v in output.items():
        if v['resource_name'] == resource_name:
            found_items[k] = v
    return found_items


def _compare_external_resources(expected_resources):
    from syndicate.core import PROCESSOR_FACADE
    compare_funcs = PROCESSOR_FACADE.compare_meta_handlers()

    errors = {}

    for resource_name, resource_meta in expected_resources.items():
        func = compare_funcs[resource_meta.get('resource_type')]
        resource_errors = func(resource_name, resource_meta)
        if resource_errors:
            errors[resource_name] = resource_errors

    if errors:
        import os
        error = f'{os.linesep}'.join(errors.values())
        raise AssertionError(error)


@exit_on_exception
def create_deployment_resources(deploy_name, bundle_name,
                                deploy_only_resources=None,
                                deploy_only_types=None,
                                excluded_resources=None,
                                excluded_types=None,
                                replace_output=False,
                                rollback_on_error=False):

    from syndicate.core import PROJECT_STATE
    latest_deploy_name = PROJECT_STATE.latest_deploy.get('deploy_name')
    latest_bundle_name = PROJECT_STATE.latest_deploy.get('bundle_name')
    latest_deploy_status = PROJECT_STATE.latest_deploy.get('operation_status')

    if latest_deploy_status is False:
        message = (f"Your last deployment with deploy name: "
                   f"{latest_deploy_name} and deploy bundle: "
                   f"{latest_bundle_name} failed, you can either clean "
                   f"resources with `syndicate clean` or continue deployment"
                   f" with `syndicate deploy --continue_deploy`.")
        raise AssertionError(message)

    latest_deploy_output = load_latest_deploy_output()
    _LOG.debug(f'Latest deploy output:\n {latest_deploy_output}')

    resources = load_meta_resources(bundle_name)
    # validate_deployment_packages(resources)
    _LOG.debug('{0} file was loaded successfully'.format(BUILD_META_FILE_NAME))

    resources = resolve_meta(resources)
    _LOG.debug('Names were resolved')
    resources = populate_s3_paths(resources, bundle_name)
    _LOG.debug('Artifacts s3 paths were resolved')

    deploy_only_resources = _resolve_names(deploy_only_resources)
    excluded_resources = _resolve_names(excluded_resources)
    _LOG.info(
        'Prefixes and suffixes of any resource names have been resolved.')

    expected_external_resources = {key: value for key, value in
                                   resources.items() if value.get('external')}
    if expected_external_resources:
        _compare_external_resources(expected_external_resources)
        _LOG.info('External resources were matched successfully')

    resources = _filter_resources(
        resources_meta=resources,
        resource_names=deploy_only_resources,
        resource_types=deploy_only_types,
        exclude_names=excluded_resources,
        exclude_types=excluded_types
    )

    _LOG.debug('Going to create: {0}'.format(prettify_json(resources)))

    # sort resources with priority
    resources_list = list(resources.items())
    resources_list.sort(key=cmp_to_key(compare_deploy_resources))

    _LOG.info('Going to deploy AWS resources')
    success, output = deploy_resources(resources_list)

    if not success:
        if rollback_on_error is True:
            USER_LOG.info(
                "Deployment failed, `rollback_on_error` is enabled,"
                " going to clean resources that have been deployed during"
                " deployment process.")
            output_resources_list = list(output.items())
            output_resources_list.sort(
                key=cmp_to_key(_compare_clean_resources))

            if latest_deploy_output:
                deploy_output_names = list(latest_deploy_output.keys())
                rollback_resources_list =\
                    [resource for resource in output_resources_list if
                     resource[0] not in deploy_output_names]
                clean_resources(rollback_resources_list)
            else:
                clean_resources(output_resources_list)

        else:
            USER_LOG.info('Going to create deploy output')
            create_deploy_output(bundle_name=bundle_name,
                                 deploy_name=deploy_name,
                                 output={**output, **latest_deploy_output},
                                 success=success,
                                 replace_output=replace_output)
    else:
        USER_LOG.info('AWS resources were deployed successfully')

        # apply dynamic changes that uses ARNs
        _LOG.info('Going to apply dynamic changes')
        _apply_dynamic_changes(resources, output)
        USER_LOG.info('Dynamic changes were applied successfully')

        _LOG.info('Going to apply common tags')
        _apply_tags(output)

        USER_LOG.info('Going to create deploy output')
        create_deploy_output(bundle_name=bundle_name,
                             deploy_name=deploy_name,
                             output={**output, **latest_deploy_output},
                             success=success,
                             replace_output=replace_output)

    if not (success is False and rollback_on_error is True):
        USER_LOG.info('Deploy output for {0} was created.'.format(deploy_name))
    return success


@exit_on_exception
def update_deployment_resources(bundle_name, deploy_name, replace_output=False,
                                update_only_types=None,
                                update_only_resources=None,
                                excluded_resources=None,
                                excluded_types=None):
    from syndicate.core import PROCESSOR_FACADE
    resources = load_meta_resources(bundle_name)
    _LOG.debug(prettify_json(resources))

    resources = resolve_meta(resources)
    _LOG.debug('Names were resolved')
    resources = populate_s3_paths(resources, bundle_name)
    _LOG.debug('Artifacts s3 paths were resolved')

    _LOG.warn(
        'Please pay attention that only the '
        'following resources types are supported for update: {}'.format(
            list(PROCESSOR_FACADE.update_handlers().keys())))

    update_only_resources = _resolve_names(update_only_resources)
    _LOG.info(
        'Prefixes and suffixes of any resource names have been resolved.')

    resources = dict((k, v) for (k, v) in resources.items() if
                     v['resource_type'] in
                     PROCESSOR_FACADE.update_handlers().keys())

    resources = _filter_resources(
        resources_meta=resources,
        resource_names=update_only_resources,
        resource_types=update_only_types,
        exclude_names=excluded_resources,
        exclude_types=excluded_types
    )

    _LOG.debug('Going to update the following resources: {0}'.format(
        prettify_json(resources)))
    resources_list = list(resources.items())
    resources_list.sort(key=cmp_to_key(_compare_update_resources))
    success, output = _process_resources(
        resources=resources_list,
        handlers_mapping=PROCESSOR_FACADE.update_handlers(),
        pass_context=True)
    create_deploy_output(bundle_name=bundle_name,
                         deploy_name=deploy_name,
                         output=output,
                         success=success,
                         replace_output=replace_output)
    return success


@exit_on_exception
def remove_deployment_resources(deploy_name, bundle_name,
                                clean_only_resources=None,
                                clean_only_types=None,
                                excluded_resources=None,
                                excluded_types=None,
                                clean_externals=None,
                                preserve_state=None):

    is_regular_output = True
    try:
        output = new_output = load_deploy_output(bundle_name, deploy_name)
        _LOG.info('Output file was loaded successfully')
    except AssertionError:
        try:
            output = new_output = load_failed_deploy_output(
                bundle_name, deploy_name
            )
            is_regular_output = False
        except AssertionError as e:
            return e

    clean_only_resources = _resolve_names(clean_only_resources)
    excluded_resources = _resolve_names(excluded_resources)
    _LOG.info(
        'Prefixes and suffixes of any resource names have been resolved.')

    if clean_externals:
        new_output = {
            k: v for k, v in new_output.items() if
            v['resource_meta'].get('external')
        }

    new_output = _filter_resources(
        resources_meta=new_output,
        resources_meta_type=DEPLOYMENT_OUTPUT,
        resource_names=clean_only_resources,
        resource_types=clean_only_types,
        exclude_names=excluded_resources,
        exclude_types=excluded_types
    )
    # sort resources with priority
    resources_list = list(new_output.items())
    resources_list.sort(key=cmp_to_key(_compare_clean_resources))
    _LOG.debug(f'Resources to delete: {prettify_json(resources_list)}')
    USER_LOG.info('Going to clean AWS resources')
    clean_resources(resources_list)
    # remove new_output from bucket
    return _post_remove_output_handling(
        deploy_name=deploy_name,
        bundle_name=bundle_name,
        preserve_state=preserve_state,
        output=output,
        new_output=new_output,
        is_regular_output=is_regular_output
    )


@exit_on_exception
def continue_deployment_resources(deploy_name, bundle_name,
                                  deploy_only_resources=None,
                                  deploy_only_types=None,
                                  excluded_resources=None,
                                  excluded_types=None,
                                  replace_output=False):
    from syndicate.core import PROJECT_STATE
    latest_deploy_name = PROJECT_STATE.latest_deploy.get('deploy_name')
    latest_bundle_name = PROJECT_STATE.latest_deploy.get('bundle_name')
    latest_deploy_status = PROJECT_STATE.latest_deploy.get('operation_status')

    if latest_deploy_status is True:
        raise AssertionError(f"Your last deployment with deploy name:"
                             f" {latest_deploy_name} and deploy bundle:"
                             f" {latest_bundle_name} was successful,"
                             f" you cannot execute the `continue deploy` "
                             f"command.")

    failed_output = load_failed_deploy_output(
        bundle_name=latest_bundle_name, deploy_name=latest_deploy_name
    )
    _LOG.info('Failed output file was loaded successfully')

    resources = load_meta_resources(bundle_name)
    _LOG.debug('{0} file was loaded successfully'.format(BUILD_META_FILE_NAME))

    resources = resolve_meta(resources)
    _LOG.debug('Names were resolved')
    resources = populate_s3_paths(resources, bundle_name)
    _LOG.debug('Artifacts s3 paths were resolved')

    deploy_only_resources = _resolve_names(deploy_only_resources)
    excluded_resources = _resolve_names(excluded_resources)
    _LOG.info(
        'Prefixes and suffixes of any resource names have been resolved.')

    resources = _filter_resources(
        resources_meta=resources,
        resource_names=deploy_only_resources,
        resource_types=deploy_only_types,
        exclude_names=excluded_resources,
        exclude_types=excluded_types
    )

    _LOG.debug('Going to create: {0}'.format(prettify_json(resources)))

    # sort resources with priority
    resources_list = list(resources.items())
    resources_list.sort(key=cmp_to_key(compare_deploy_resources))

    success, updated_output = continue_deploy_resources(
        resources=resources_list, failed_output=failed_output)
    _LOG.info('AWS resources were deployed successfully')
    if success:
        # apply dynamic changes that uses ARNs
        _LOG.info('Going to apply dynamic changes')
        _apply_dynamic_changes(resources, updated_output)
        _LOG.info('Dynamic changes were applied successfully')

        _LOG.info('Going to apply common tags')
        _apply_tags(updated_output)

    # remove failed output from bucket
    remove_failed_deploy_output(bundle_name, deploy_name)
    _LOG.info('Going to create deploy output')
    create_deploy_output(bundle_name=bundle_name,
                         deploy_name=deploy_name,
                         output=updated_output,
                         success=success,
                         replace_output=replace_output)
    return success


def _post_remove_output_handling(deploy_name, bundle_name, preserve_state,
                                 output, new_output, is_regular_output):
    if output == new_output:
        if not preserve_state:
            # remove output from bucket
            remove_failed_deploy_output(bundle_name, deploy_name)
    else:
        for key, value in new_output.items():
            output.pop(key)
        create_deploy_output(bundle_name=bundle_name,
                             deploy_name=deploy_name,
                             output=output,
                             success=is_regular_output,
                             replace_output=True)
        return {'operation': PARTIAL_CLEAN_ACTION}


def _apply_dynamic_changes(resources, output):
    from syndicate.core import PROCESSOR_FACADE
    pool = ThreadPoolExecutor(max_workers=5)
    futures = []
    for name, meta in resources.items():
        resource_type = meta['resource_type']
        apply_changes = meta.get('apply_changes')
        if apply_changes:
            for apply_item in apply_changes:
                change_type = apply_item['apply_type']
                dependency_name = apply_item['dependency_name']
                res_config = resources.get(dependency_name)
                if not res_config:
                    _LOG.debug('Dependency resource {0} is not found, '
                               'skipping the apply'.format(dependency_name))
                else:
                    dependency_type = res_config['resource_type']
                    func = PROCESSOR_FACADE.resource_identifier() \
                        .get(resource_type)
                    if func:
                        resource_output = __find_output_by_resource_name(
                            output, name)
                        identifier = func(name, resource_output)
                        apply_func = PROCESSOR_FACADE.mapping_applier() \
                            .get(change_type)
                        if apply_func:
                            alias = '#{' + name + '}'
                            f = pool.submit(apply_func, alias, identifier,
                                            apply_item)
                            futures.append(f)
                        else:
                            _LOG.warn('Dynamic apply is not defined '
                                      'for {0} type'.format(change_type))
                    else:
                        _LOG.warn('Resource identifier is not defined '
                                  'for {0} type'.format(dependency_type))
            _LOG.info('Dynamic changes were applied to {0}'.format(name))
    concurrent.futures.wait(futures, timeout=None, return_when=ALL_COMPLETED)


def _apply_tags(output: dict):
    from syndicate.core import RESOURCES_PROVIDER
    tags_resource = RESOURCES_PROVIDER.tags_api()
    tags_resource.apply_tags(output)


def compare_deploy_resources(first, second):
    first_resource_type = first[-1]['resource_type']
    second_resource_type = second[-1]['resource_type']
    first_res_priority = DEPLOY_RESOURCE_TYPE_PRIORITY[first_resource_type]
    second_res_priority = DEPLOY_RESOURCE_TYPE_PRIORITY[second_resource_type]
    return _compare_res(first_res_priority, second_res_priority)


def _compare_clean_resources(first, second):
    first_resource_type = first[-1]['resource_meta']['resource_type']
    second_resource_type = second[-1]['resource_meta']['resource_type']
    first_res_priority = CLEAN_RESOURCE_TYPE_PRIORITY[first_resource_type]
    second_res_priority = CLEAN_RESOURCE_TYPE_PRIORITY[second_resource_type]
    return _compare_res(first_res_priority, second_res_priority)


def _compare_update_resources(first, second):
    first_resource_type = first[-1]['resource_type']
    second_resource_type = second[-1]['resource_type']
    first_res_priority = UPDATE_RESOURCE_TYPE_PRIORITY[first_resource_type]
    second_res_priority = UPDATE_RESOURCE_TYPE_PRIORITY[second_resource_type]
    return _compare_res(first_res_priority, second_res_priority)


def _compare_res(first_res_priority, second_res_priority):
    if first_res_priority < second_res_priority:
        return -1
    elif first_res_priority > second_res_priority:
        return 1
    else:
        return 0


def _resolve_names(names):
    from syndicate.core import CONFIG
    preset_name_resolution = functools.partial(resolve_resource_name,
                                               prefix=CONFIG.resources_prefix,
                                               suffix=CONFIG.resources_suffix)
    resolve_n_unify_names = lambda collection: set(
        collection + tuple(map(preset_name_resolution, collection)))

    return resolve_n_unify_names(names or tuple())


def _filter_resources(resources_meta, resources_meta_type=BUILD_META,
                      resource_names=None, resource_types=None,
                      exclude_names=None, exclude_types=None):
    resource_names = set() if resource_names is None else set(resource_names)
    resource_types = set() if resource_types is None else set(resource_types)
    exclude_names = set() if exclude_names is None else set(exclude_names)
    exclude_types = set() if exclude_types is None else set(exclude_types)

    _LOG.debug(f"Include resources by name: {list(resource_names) or 'All'}")
    _LOG.debug(f"Include resources by type: {list(resource_types) or 'All'}")
    _LOG.debug(f"Exclude resources by name: {list(exclude_names) or 'None'}")
    _LOG.debug(f"Exclude resources by type: {list(exclude_types) or 'None'}")

    if not any([resource_names, resource_types]):
        filtered = resources_meta
    else:
        if resources_meta_type == BUILD_META:
            filtered = {
                k: v for k, v in resources_meta.items()
                if k in resource_names or v['resource_type'] in resource_types
            }
        elif resources_meta_type == DEPLOYMENT_OUTPUT:
            filtered = {
                k: v for k, v in resources_meta.items()
                if v['resource_name'] in resource_names
                or v['resource_meta']['resource_type'] in resource_types
            }

    for k, v, in copy.deepcopy(filtered).items():
        if resources_meta_type == BUILD_META:
            if k in exclude_names or v['resource_type'] in exclude_types:
                filtered.pop(k)
        elif resources_meta_type == DEPLOYMENT_OUTPUT:
            if (v['resource_name'] in exclude_names
                    or v['resource_meta']['resource_type'] in exclude_types):
                filtered.pop(k)

    return filtered



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import logging
import getpass
import os
import sys
from pathlib import Path
from logging import DEBUG, Formatter, INFO, getLogger
from datetime import date

LOG_USER_HOME_FOLDER_NAME = '.syndicate_logs'
LOG_FOLDER_NAME = 'logs'
LOG_FILE_NAME = '%Y-%m-%d-syndicate.log'
LOG_NAME = 'syndicate'
LOG_LEVEL = (DEBUG
             if os.environ.get('SDCT_DEBUG', '').lower() == 'true'
             else INFO)
USER_NAME = getpass.getuser()
LOG_FORMAT_FOR_FILE = ('%(asctime)s [%(levelname)s] USER:{} %(filename)s:'
                       '%(lineno)d:%(funcName)s LOG: %(message)s'
                       .format(USER_NAME))


def get_project_log_file_path() -> str:
    """Returns the path to the file where logs will be saved.
    :rtype: str
    :returns: a path to the main log file
    """
    sdct_conf = os.getenv('SDCT_CONF')
    if sdct_conf:
        logs_path = os.path.join(sdct_conf, LOG_FOLDER_NAME)
    else:
        logs_folder_path = os.environ.get('SDCT_LOGS') or Path.home()
        logs_path = os.path.join(logs_folder_path, LOG_USER_HOME_FOLDER_NAME)

    try:
        os.makedirs(logs_path, exist_ok=True)
    except OSError as e:
        print(f'Error while creating logs path: {e}', file=sys.stderr)

    today = date.today()
    log_file_path = os.path.join(logs_path, today.strftime(LOG_FILE_NAME))

    return log_file_path


log_file_path = get_project_log_file_path()

# formatter
formatter = Formatter(LOG_FORMAT_FOR_FILE)
# file output
file_handler = logging.FileHandler(filename=log_file_path)
file_handler.setFormatter(formatter)
# console output
console_handler = logging.StreamHandler()
console_handler.setFormatter(formatter)

user_logger = getLogger(f'user-{LOG_NAME}')
user_logger.addHandler(console_handler)
user_logger.addHandler(file_handler)

syndicate_logger = getLogger(LOG_NAME)
syndicate_logger.addHandler(file_handler)

if LOG_LEVEL == DEBUG:
    syndicate_logger.addHandler(console_handler)

logging.captureWarnings(True)


def get_logger(log_name, level=LOG_LEVEL):
    """
    :param level:   CRITICAL = 50
                    ERROR = 40
                    WARNING = 30
                    INFO = 20
                    DEBUG = 10
                    NOTSET = 0
    :type log_name: str
    :type level: int
    """
    module_logger = syndicate_logger.getChild(log_name)
    if level:
        module_logger.setLevel(level)
    return module_logger


def get_user_logger(level=LOG_LEVEL):
    """
    :param level:   CRITICAL = 50
                    ERROR = 40
                    WARNING = 30
                    INFO = 20
                    DEBUG = 10
                    NOTSET = 0
    :type level: int
    """
    module_logger = user_logger.getChild('child')
    if level:
        module_logger.setLevel(level)
    return module_logger



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from troposphere import batch

from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import dict_keys_to_upper_camel_case
from syndicate.core.resources.batch_compenv_resource import \
    (DEFAULT_STATE,
     DEFAULT_SERVICE_ROLE)
from .cf_iam_role_converter import CfIamRoleConverter
from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import (iam_role_logic_name,
                                  iam_instance_profile_logic_name,
                                  batch_compute_env_logic_name)

_LOG = get_logger('cf_batch_compenv_converter')


class CfBatchComputeEnvironmentConverter(CfResourceConverter):

    def convert(self, name, meta):
        logic_name = batch_compute_env_logic_name(name)
        comp_env = batch.ComputeEnvironment(logic_name)

        state = meta.get('state')
        if not state:
            state = DEFAULT_STATE

        iam_conn = self.resources_provider.iam().iam_conn
        role_name = meta.get('service_role')
        if not role_name:
            role_name = DEFAULT_SERVICE_ROLE
            default_role = iam_conn.get_role(role_name=role_name)
            if not default_role:
                default_role = self.get_resource(iam_role_logic_name(role_name))

            if not default_role:
                _LOG.warn("Default Service Role '%s' not found "
                          "and will be added to the template.", role_name)
                allowed_account = iam_conn.resource.CurrentUser().arn.split(':')[4]
                role_converter = CfIamRoleConverter(
                    template=self.template,
                    config=self.config,
                    resources_provider=self.resources_provider)

                role_converter.convert(role_name, {
                    'allowed_accounts': allowed_account,
                    'principal_service': 'batch'
                })
                default_role = self.get_resource(iam_role_logic_name(role_name))
                policy = iam_conn.get_policy_arn(role_name)
                if not policy:
                    raise AssertionError(
                        'IAM Policy "{}" does not exist'.format(role_name))
                CfIamRoleConverter.attach_managed_policy(role=default_role,
                                                         policy=policy)

        role = iam_conn.check_if_role_exists(role_name=role_name)
        if not role:
            role = self.get_resource(iam_role_logic_name(role_name))
            if not role:
                raise AssertionError('IAM role "{}" does not exist.'
                                     .format(role_name))
            comp_env.ServiceRole = role.get_att('Arn')
        else:
            comp_env.ServiceRole = role

        compute_res = meta.get('compute_resources')
        if compute_res:
            tags = compute_res.get('tags', {}).copy()
            instance_role = compute_res.get('instance_role')
            compute_res = dict_keys_to_upper_camel_case(compute_res)
            if tags:
                compute_res['Tags'] = tags
            if instance_role:
                instance_profile = self.get_resource(
                    iam_instance_profile_logic_name(instance_role))
                if instance_profile:
                    compute_res['InstanceRole'] = instance_profile.ref()
            comp_env.ComputeResources = batch.ComputeResources.from_dict(
                title=None, d=compute_res)

        comp_env.Type = meta['compute_environment_type']
        comp_env.ComputeEnvironmentName = name
        comp_env.State = state
        self.template.add_resource(comp_env)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import uuid
from json import dumps, loads

from boto3 import client
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('syndicate.connection.sns_connection')


@apply_methods_decorator(retry())
class SNSConnection(object):
    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('sns', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new SNS connection.')

    def create_topic(self, name):
        """ Crete SNS topic and return topic arn.

        :type name: str
        """
        return self.client.create_topic(Name=name)['TopicArn']

    def subscribe(self, endpoint, topic_name, protocol):
        """
        :param protocol:
            http -- delivery of JSON-encoded message via HTTP POST
            https -- delivery of JSON-encoded message via HTTPS POST
            email -- delivery of message via SMTP
            email-json -- delivery of JSON-encoded message via SMTP
            sms -- delivery of message via SMS
            sqs -- delivery of JSON-encoded message to an Amazon SQS queue
            application -- delivery of JSON-encoded message to an EndpointArn
                           for a mobile app and device.
            lambda -- delivery of JSON message to an AWS Lambda function
        :type protocol: str
        :type topic_name: str
        :param endpoint:
            http protocol, the endpoint is an URL beginning with "http://"
            https protocol, the endpoint is a URL beginning with "https://"
            email protocol, the endpoint is an email address
            email-json protocol, the endpoint is an email address
            sms protocol, the endpoint is a phone number of an SMS-enabled
                device
            sqs protocol, the endpoint is the ARN of an Amazon SQS queue
            application protocol, the endpoint is the EndpointArn of a mobile
                app and device.
            lambda protocol, the endpoint is the ARN of an AWS Lambda function
        :type endpoint: str
        """
        topic_arn = self.get_topic_arn(topic_name)
        if topic_arn is None:
            raise AssertionError(
                'Topic does not exist: {0}.'.format(topic_name))
        self.client.subscribe(TopicArn=topic_arn,
                              Protocol=protocol,
                              Endpoint=endpoint)
        return topic_arn

    def get_topic_arn(self, name):
        """ Get topic arn by name.

        :type name: str
        """
        topics = self.get_topics()
        for each in topics:
            if name == each['TopicArn'].split(':')[-1]:
                return each['TopicArn']

    def get_platform_application(self, name):
        """ Get application arn by name.

        :type name: str
        """
        applications = self.get_platform_applications()
        for each in applications:
            resolved_item = each['PlatformApplicationArn'].split(':')[-1]
            if name == resolved_item.split('/')[-1]:
                return each['PlatformApplicationArn']

    def is_user_subscribed(self, endpoint, topic_name):
        topic_arn = self.get_topic_arn(topic_name)
        subscriptions = self.client.list_subscriptions_by_topic(
            TopicArn=topic_arn)['Subscriptions']
        for each in subscriptions:
            if endpoint == each['Endpoint']:
                return True

    def publish_message(self, topic_name, message):
        topic_arn = self.get_topic_arn(topic_name)
        return self.client.publish(
            TargetArn=topic_arn,
            Message=message,
            MessageAttributes={
                'string': {
                    'DataType': 'String',
                    'StringValue': ' '
                }
            }
        )

    def get_topics(self):
        """ Get all topics."""
        topics = []
        response = self.client.list_topics()
        topics.extend(response.get('Topics'))
        token = response.get('NextToken')
        while token:
            response = self.client.list_topics(NextToken=token)
            topics.extend(response.get('Topics'))
            token = response.get('NextToken')
        return topics

    def get_platform_applications(self):
        """ Get all platform applications."""
        applications = []
        response = self.client.list_platform_applications()
        applications.extend(response.get('PlatformApplications'))
        token = response.get('NextToken')
        while token:
            response = self.client.list_platform_applications(NextToken=token)
            applications.extend(response.get('PlatformApplications'))
            token = response.get('NextToken')
        return applications

    def remove_topic_by_arn(self, topic_arn):
        """ Remove topic by arn.

        :type topic_arn: str
        """
        self.client.delete_topic(TopicArn=topic_arn)

    def remove_topic_by_name(self, topic_name):
        """ Remove topic by arn.

        :type topic_name: str
        """
        arn = self.get_topic_arn(topic_name)
        if arn:
            self.client.delete_topic(TopicArn=arn)

    def set_topic_attribute(self, topic_arn, attr_name, attr_value):
        self.client.set_topic_attributes(
            TopicArn=topic_arn,
            AttributeName=attr_name,
            AttributeValue=attr_value
        )

    def allow_service_invoke(self, topic_arn, service):
        existing_attr = self.get_topic_attributes(topic_arn)
        existing_policy = existing_attr['Attributes']['Policy']
        existing_policy_dict = loads(existing_policy)
        policy = {
            "Sid": str(uuid.uuid1()),
            "Effect": "Allow",
            "Principal":
                {
                    "Service": "{0}".format(service)
                },
            "Action": "sns:Publish",
            "Resource": "{0}".format(topic_arn)
        }
        existing_policy_dict['Statement'].append(policy)
        self.set_topic_attribute(topic_arn, 'Policy',
                                 dumps(existing_policy_dict))

    def get_topic_attributes(self, topic_arn):
        return self.client.get_topic_attributes(
            TopicArn=topic_arn
        )

    def get_platform_application_attributes(self, application_arn):
        return self.client.get_platform_application_attributes(
            PlatformApplicationArn=application_arn
        )

    def add_account_permission(self, topic_arn, account_id, action, label):
        if isinstance(account_id, str):
            account_id = [account_id]
        if not isinstance(account_id, list):
            raise AssertionError('Incorrect account id {0}'.format(account_id))

        if isinstance(action, str):
            action = [action]
        if not isinstance(action, list):
            raise AssertionError('Incorrect action {0}'.format(action))

        self.client.add_permission(TopicArn=topic_arn, Label=label,
                                   AWSAccountId=account_id,
                                   ActionName=action)

    def revoke_account_permission(self, topic_arn, label):
        self.client.remove_permission(TopicArn=topic_arn, Label=label)

    def list_subscriptions_by_topic(self, topic_arn):
        subscriptions = []
        try:
            response = self.client.list_subscriptions_by_topic(
                TopicArn=topic_arn)
        except ClientError as e:
            if e.response['Error']['Code'] == 'NotFound':
                _LOG.warn(f'SNS topic \'{topic_arn}\' is not found')
                return subscriptions
            else:
                raise e
        subscriptions.extend(response.get('Subscriptions'))
        token = response.get('NextToken')
        while token:
            response = self.client.list_subscriptions_by_topic(
                TopicArn=topic_arn, NextToken=token)
            subscriptions.extend(response.get('Subscriptions'))
            token = response.get('NextToken')

        return subscriptions

    def unsubscribe(self, subscription_arn):
        self.client.unsubscribe(SubscriptionArn=subscription_arn)

    def create_platform_endpoint(self, platform_application_arn, token):
        response = self.client.create_platform_endpoint(
            PlatformApplicationArn=platform_application_arn,
            Token=token
        )
        return response.get('EndpointArn')

    def create_platform_application(self, name, platform, attributes):
        response = self.client.create_platform_application(
            Name=name, Platform=platform, Attributes=attributes)
        return response.get('PlatformApplicationArn')

    def remove_application_by_arn(self, application_arn):
        """ Remove application by arn.

        :type application_arn: str
        """
        self.client.delete_platform_application(
            PlatformApplicationArn=application_arn)



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json

from syndicate.core.transform.terraform.converter.tf_resource_converter import \
    TerraformResourceConverter


class S3BucketConverter(TerraformResourceConverter):

    def convert(self, name, resource):
        acl = resource.get('acl')
        policy = resource.get('policy')

        rule_document = resource.get('LifecycleConfiguration')
        rules = []
        if rule_document:
            for rule in rule_document['Rules']:
                if 'Prefix' not in rule:
                    rule['Prefix'] = ''
                rules.append(rule)

        cors_configuration = resource.get('cors')
        s3_bucket_meta = s3_bucket(bucket_name=name, acl=acl,
                                   policy=json.dumps(policy),
                                   cors_rules=cors_configuration,
                                   lifecycle_rules=rules)
        self.template.add_aws_s3_bucket(meta=s3_bucket_meta)


def s3_bucket(bucket_name, acl, policy, cors_rules=None, lifecycle_rules=None):
    s3_bucket = {
        "acl": acl,
        "bucket": bucket_name
    }

    if policy:
        s3_bucket['policy'] = policy

    cors = []
    if cors_rules:
        for rule in cors_rules:
            cors.append(
                {
                    'allowed_headers': rule.get('AllowedHeaders'),
                    'allowed_methods': rule.get('AllowedMethods'),
                    'allowed_origins': rule.get('AllowedOrigins'),
                    'expose_headers': rule.get('ExposeHeaders'),
                    'max_age_seconds': rule.get('MaxAgeSeconds')
                }
            )
    if cors:
        s3_bucket['cors_rule'] = cors

    rules = []
    if lifecycle_rules:
        for rule in lifecycle_rules:
            transitions = []
            transitions_list = rule.get('Transitions')
            if transitions_list:
                for transition in transitions_list:
                    transitions.append(
                        {
                            'days': int(transition['Days']),
                            'storage_class': transition['StorageClass']
                        }
                    )
            rules.append(
                {
                    'id': rule['ID'],
                    'prefix': rule['Prefix'],
                    'enabled': True if rule['Status'] == 'Enabled' else False,
                    'expiration': {
                        'days': int(rule['Expiration']['Days'])
                    },
                    'transition': transitions
                }
            )
    if rules:
        s3_bucket['lifecycle_rule'] = rules

    resource = {
        bucket_name: s3_bucket

    }
    return resource



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from boto3 import client

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('syndicate.connection.ses_connection')


@apply_methods_decorator(retry())
class SESConnection(object):
    """SES connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, source_arn=None,
                 aws_session_token=None):
        self.client = client('ses', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        self.charset = 'utf-8'
        self.arn = source_arn
        self.source_email = source_arn[source_arn.index('/') + 1:]
        _LOG.debug('Opened new SES connection.')

    def send_email_from_identity(self, to_addresses, cc_addresses=None,
                                 bcc_addresses=None, reply_to_addresses=None,
                                 subject='', body_type='Html', body='body'):
        """
        Method for sending message from specified arn in AWS SES.
        :type to_addresses: list
        :type cc_addresses: list
        :type bcc_addresses: list
        :type reply_to_addresses: list
        :type subject: str
        :param body_type: 'Html'/'Text'
        :type body: str
        """
        cc_addresses = cc_addresses if cc_addresses else []
        bcc_addresses = bcc_addresses if bcc_addresses else []
        reply_to_addresses = reply_to_addresses if reply_to_addresses else []
        response = self.client.send_email(
            Source=self.source_email,
            Destination={
                'ToAddresses': to_addresses,
                'CcAddresses': cc_addresses,
                'BccAddresses': bcc_addresses
            },
            Message={
                'Subject': {
                    'Data': subject,
                    'Charset': self.charset
                },
                'Body': {
                    body_type: {
                        'Data': body,
                        'Charset': self.charset
                    }
                }
            },
            ReplyToAddresses=reply_to_addresses,
            ReturnPath=self.source_email,
            SourceArn=self.arn,
            ReturnPathArn=self.arn
        )
        return response



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
from json import load
from typing import Any
from urllib.parse import urlparse
from syndicate.commons.log_helper import get_logger
from syndicate.core.build.helper import (build_py_package_name,
                                         resolve_bundle_directory)
from syndicate.core.build.validator.mapping import (VALIDATOR_BY_TYPE_MAPPING,
                                                    ALL_TYPES)
from syndicate.core.conf.processor import GLOBAL_AWS_SERVICES
from syndicate.core.constants import (API_GATEWAY_TYPE, ARTIFACTS_FOLDER,
                                      BUILD_META_FILE_NAME, EBS_TYPE,
                                      LAMBDA_CONFIG_FILE_NAME, LAMBDA_TYPE,
                                      RESOURCES_FILE_NAME, RESOURCE_LIST,
                                      IAM_ROLE, LAMBDA_LAYER_TYPE,
                                      S3_PATH_NAME,
                                      LAMBDA_LAYER_CONFIG_FILE_NAME,
                                      WEB_SOCKET_API_GATEWAY_TYPE,
                                      OAS_V3_FILE_NAME,
                                      API_GATEWAY_OAS_V3_TYPE, SWAGGER_UI_TYPE,
                                      SWAGGER_UI_CONFIG_FILE_NAME)
from syndicate.core.helper import (build_path, prettify_json,
                                   resolve_aliases_for_string,
                                   write_content_to_file)
from syndicate.core.resources.helper import resolve_dynamic_identifier

DEFAULT_IAM_SUFFIX_LENGTH = 5

_LOG = get_logger('syndicate.core.build.meta_processor')


def validate_deployment_packages(bundle_path, meta_resources):
    package_paths = artifact_paths(meta_resources)
    nonexistent_packages = []
    for package in package_paths:
        package_path = build_path(bundle_path, package)
        if not os.path.exists(package_path):
            nonexistent_packages.append(package_path)

    if nonexistent_packages:
        raise AssertionError('Bundle is not properly configured.'
                             ' Nonexistent deployment packages: '
                             '{0}'.format(prettify_json(nonexistent_packages)))


def artifact_paths(meta_resources):
    return [i for i in
            [_retrieve_package(v) for v in list(meta_resources.values())] if i]


def _retrieve_package(meta):
    s3_path = meta.get(S3_PATH_NAME)
    if s3_path:
        return s3_path


def _check_duplicated_resources(initial_meta_dict, additional_item_name,
                                additional_item):
    """ Match two meta dicts (overall and separated) for duplicates.

    :type initial_meta_dict: dict
    :type additional_item_name: str
    :type additional_item: dict
    """
    if additional_item_name in initial_meta_dict:
        additional_type = additional_item['resource_type']
        initial_item = initial_meta_dict.get(additional_item_name)
        if not initial_item:
            return
        initial_type = initial_item['resource_type']
        if additional_type == initial_type and initial_type in \
                {API_GATEWAY_TYPE, WEB_SOCKET_API_GATEWAY_TYPE}:
            # check if APIs have same resources
            for each in list(initial_item['resources'].keys()):
                if each in list(additional_item['resources'].keys()):
                    raise AssertionError(
                        "API '{0}' has duplicated resource '{1}'! Please, "
                        "change name of one resource or remove one.".format(
                            additional_item_name, each))
            # check if APIs have duplicated cluster configurations
            for config in ['cluster_cache_configuration',
                           'cluster_throttling_configuration']:
                initial_config = initial_item.get(config)
                additional_config = additional_item.get(config)
                if initial_config and additional_config:
                    raise AssertionError(
                        "API '{0}' has duplicated {1}. Please, remove one "
                        "configuration.".format(additional_item_name, config)
                    )
                if initial_config:
                    additional_item[config] = initial_config
            # handle responses
            initial_responses = initial_item.get(
                'api_method_responses')
            additional_responses = additional_item.get(
                'api_method_responses')
            if initial_responses and additional_responses:
                raise AssertionError(
                    "API '{0}' has duplicated api method responses "
                    "configurations. Please, remove one "
                    "api method responses configuration.".format(
                        additional_item_name)
                )
            if initial_responses:
                additional_item[
                    'api_method_responses'] = initial_responses
            # handle integration responses
            initial_integration_resp = initial_item.get(
                'api_method_integration_responses')
            additional_integration_resp = additional_item.get(
                'api_method_integration_responses')
            if initial_integration_resp and additional_integration_resp:
                raise AssertionError(
                    "API '{0}' has duplicated api method integration "
                    "responses configurations. Please, remove one "
                    "api method integration responses configuration.".format(
                        additional_item_name)
                )
            if initial_integration_resp:
                additional_item[
                    'api_method_integration_responses'] = initial_integration_resp
            # join items dependencies
            dependencies_dict = {each['resource_name']: each
                                 for each in
                                 additional_item.get('dependencies') or []}
            for each in initial_item.get('dependencies') or []:
                if each['resource_name'] not in dependencies_dict:
                    additional_item['dependencies'].append(each)
            # join items resources
            additional_item['resources'].update(initial_item['resources'])
            # return aggregated API description
            init_deploy_stage = initial_item.get('deploy_stage')
            if init_deploy_stage:
                additional_item['deploy_stage'] = init_deploy_stage

            init_compression = initial_item.get("minimum_compression_size")
            if init_compression:
                additional_comp_size = \
                    additional_item.get('minimum_compression_size')
                if additional_comp_size:
                    _LOG.warn(f"Found 'minimum_compression_size': "
                              f"{init_compression} inside root "
                              f"deployment_resources. The value "
                              f"'{additional_comp_size}' from: "
                              f"{additional_item} will be overwritten")
                additional_item['minimum_compression_size'] = init_compression

            # join authorizers
            initial_authorizers = initial_item.get('authorizers') or {}
            additional_authorizers = additional_item.get('authorizers') or {}
            additional_item['authorizers'] = {**initial_authorizers,
                                              **additional_authorizers}
            # join models
            initial_models = initial_item.get('models') or {}
            additional_models = additional_item.get('models') or {}
            additional_item['models'] = {**initial_models, **additional_models}
            # policy statement singleton
            _pst = initial_item.get('policy_statement_singleton')
            if 'policy_statement_singleton' not in additional_item and _pst:
                additional_item['policy_statement_singleton'] = _pst

            additional_item['route_selection_expression'] = initial_item.get(
                'route_selection_expression')

            additional_item = _merge_api_gw_list_typed_configurations(
                initial_item,
                additional_item,
                ['binary_media_types', 'apply_changes']
            )

            return additional_item

        else:
            initial_item_type = initial_item.get("resource_type")
            additional_item_type = additional_item.get("resource_type")
            raise AssertionError(
                f"Two resources with equal names were found! "
                f"Name: '{additional_item_name}', first resource type: "
                f"'{initial_item_type}', second resource type: "
                f"'{additional_item_type}'. \nPlease, rename one of them!"
            )


def _merge_api_gw_list_typed_configurations(initial_resource,
                                            additional_resource,
                                            property_names_list):
    for property_name in property_names_list:
        initial_property_value = initial_resource.get(property_name, [])
        additional_resource_value = additional_resource.get(property_name, [])
        additional_resource[
            property_name] = initial_property_value + additional_resource_value
    return additional_resource


def _populate_s3_path_python_node(meta, bundle_name):
    name = meta.get('name')
    version = meta.get('version')
    prefix = meta.pop('prefix', None)
    suffix = meta.pop('suffix', None)
    if not name or not version:
        raise AssertionError('Lambda config must contain name and version. '
                             'Existing configuration'
                             ': {0}'.format(prettify_json(meta)))
    else:
        if prefix:
            name = name[len(prefix):]
        if suffix:
            name = name[:-len(suffix)]
        meta[S3_PATH_NAME] = build_path(bundle_name,
                                        build_py_package_name(name, version))


def _populate_s3_path_java(meta, bundle_name):
    deployment_package = meta.get('deployment_package')
    if not deployment_package:
        raise AssertionError('Lambda config must contain deployment_package. '
                             'Existing configuration'
                             ': {0}'.format(prettify_json(meta)))
    else:
        meta[S3_PATH_NAME] = build_path(bundle_name, deployment_package)


def _populate_s3_path_lambda(meta, bundle_name):
    runtime = meta.get('runtime')
    if not runtime:
        raise AssertionError(
            'Lambda config must contain runtime. '
            'Existing configuration: {0}'.format(prettify_json(meta)))
    resolver_func = RUNTIME_PATH_RESOLVER.get(runtime.lower())
    if resolver_func:
        resolver_func(meta, bundle_name)
    else:
        raise AssertionError(
            'Specified runtime {0} in {1} is not supported. '
            'Supported runtimes: {2}'.format(
                runtime.lower(), meta.get('name'),
                list(RUNTIME_PATH_RESOLVER.keys())))


def _populate_s3_path_lambda_layer(meta, bundle_name):
    deployment_package = meta.get('deployment_package')
    if not deployment_package:
        raise AssertionError(
            'Lambda Layer config must contain deployment_package. '
            'Existing configuration'
            ': {0}'.format(prettify_json(meta)))
    else:
        meta[S3_PATH_NAME] = build_path(bundle_name, deployment_package)


def _populate_s3_path_ebs(meta, bundle_name):
    deployment_package = meta.get('deployment_package')
    if not deployment_package:
        raise AssertionError('Beanstalk_app config must contain '
                             'deployment_package. Existing configuration'
                             ': {0}'.format(prettify_json(meta)))
    else:
        meta[S3_PATH_NAME] = build_path(bundle_name, deployment_package)


def _populate_s3_path_swagger_ui(meta, bundle_name):
    deployment_package = meta.get('deployment_package')
    if not deployment_package:
        raise AssertionError('Swagger UI config must contain '
                             'deployment_package. Existing configuration'
                             ': {0}'.format(prettify_json(meta)))
    else:
        meta[S3_PATH_NAME] = build_path(bundle_name, deployment_package)


def populate_s3_paths(overall_meta, bundle_name):
    for name, meta in overall_meta.items():
        resource_type = meta.get('resource_type')
        mapping_func = S3_PATH_MAPPING.get(resource_type)
        if mapping_func:
            mapping_func(meta, bundle_name)
    return overall_meta


def extract_deploy_stage_from_openapi_spec(openapi_spec: dict) -> str:
    """
    Extract the first path segment from the server URL in an API specification.
    If no server URL is found, or there is no path segment, raise an exception.
    """

    servers = openapi_spec.get('servers', [])
    if not servers:
        raise ValueError("No server information found in API specification.")

    server_url = servers[0].get('url', '')
    variables = servers[0].get('variables', {})

    # Substitute variables in the URL template with their default values, if any
    for var_name, var_details in variables.items():
        default_value = var_details.get('default', '')
        server_url = server_url.replace(f'{{{var_name}}}', default_value)

    # Extract the first path segment
    path_segments = [segment for segment in
                     urlparse(server_url).path.split('/')
                     if segment]
    if not path_segments:
        raise ValueError("No path segments found in server URL.")

    return path_segments[0]


RUNTIME_PATH_RESOLVER = {
    'python3.8': _populate_s3_path_python_node,
    'python3.9': _populate_s3_path_python_node,
    'python3.10': _populate_s3_path_python_node,
    'python3.11': _populate_s3_path_python_node,
    'python3.12': _populate_s3_path_python_node,
    'java11': _populate_s3_path_java,
    'java17': _populate_s3_path_java,
    'java21': _populate_s3_path_java,
    'nodejs16.x': _populate_s3_path_python_node,
    'nodejs18.x': _populate_s3_path_python_node,
    'nodejs20.x': _populate_s3_path_python_node
}

S3_PATH_MAPPING = {
    LAMBDA_TYPE: _populate_s3_path_lambda,
    EBS_TYPE: _populate_s3_path_ebs,
    LAMBDA_LAYER_TYPE: _populate_s3_path_lambda_layer,
    SWAGGER_UI_TYPE: _populate_s3_path_swagger_ui
}


def _look_for_configs(nested_files: list[str], resources_meta: dict[str, Any],
                      path: str, bundle_name: str) -> None:
    """ Look for all config files in project structure. Read content and add
    all meta to overall meta if there is no duplicates. If duplicates found -
    raise AssertionError.

    :param nested_files: A list of files in the project
    :param resources_meta: A dictionary of resources metadata
    :param path: A string path to the project
    :param bundle_name: A string name of the bundle
    """
    for each in nested_files:
        if each.endswith(LAMBDA_CONFIG_FILE_NAME) or \
                each.endswith(LAMBDA_LAYER_CONFIG_FILE_NAME) or \
                each.endswith(SWAGGER_UI_CONFIG_FILE_NAME):
            resource_config_path = os.path.join(path, each)
            _LOG.debug(f'Processing file: {resource_config_path}')
            with open(resource_config_path) as data_file:
                resource_conf = load(data_file)

            resource_name = resource_conf['name']
            resource_type = resource_conf['resource_type']
            _LOG.debug(f'Found {resource_type}: {resource_name}')
            res = _check_duplicated_resources(resources_meta, resource_name,
                                              resource_conf)
            if res:
                resource_conf = res
            resources_meta[resource_name] = resource_conf

        if each.endswith(OAS_V3_FILE_NAME):
            openapi_spec_path = os.path.join(path, each)
            _LOG.debug(f'Processing file: {openapi_spec_path}')
            with open(openapi_spec_path) as data_file:
                openapi_spec = load(data_file)

            api_gateway_name = openapi_spec['info']['title']
            _LOG.debug(f'Found API Gateway: {api_gateway_name}')
            deploy_stage = extract_deploy_stage_from_openapi_spec(openapi_spec)
            resource = {
                "definition": openapi_spec,
                "resource_type": API_GATEWAY_OAS_V3_TYPE,
                "deploy_stage": deploy_stage
            }
            res = _check_duplicated_resources(
                resources_meta, api_gateway_name, resource
            )
            if res:
                resource = res
            resources_meta[api_gateway_name] = resource

        if each == RESOURCES_FILE_NAME:
            additional_config_path = os.path.join(path, RESOURCES_FILE_NAME)
            _LOG.debug('Processing file: {0}'.format(additional_config_path))
            with open(additional_config_path, encoding='utf-8') as json_file:
                deployment_resources = load(json_file)
            for resource_name in deployment_resources:
                _LOG.debug('Found resource ' + resource_name)
                resource = deployment_resources[resource_name]
                # check if resource type exists in deployment framework and
                #  has resource_type field
                try:
                    resource_type = resource['resource_type']
                except KeyError:
                    error_message = ("There is no 'resource_type' "
                                     "in {0}").format(resource_name)
                    _LOG.error(error_message)
                    raise AssertionError(error_message)
                if resource_type not in RESOURCE_LIST:
                    error_message = (
                        f'Unsupported resource type found: "{resource_type}". '
                        f'Please double-check the correctness of the specified '
                        f'resource type. To add a new resource type please '
                        f'request the support team.')
                    _LOG.error(error_message)
                    raise KeyError(error_message)
                res = _check_duplicated_resources(resources_meta,
                                                  resource_name, resource)
                if res:
                    resource = res
                resources_meta[resource_name] = resource


# todo validate all required configs
def create_resource_json(project_path: str, bundle_name: str) -> dict[
    str, Any]:
    """ Create resource catalog json with all resource metadata in project.

    :param project_path: path to the project
    :type bundle_name: name of the bucket subdir
    """
    resources_meta = {}

    # Walking through every folder in the project
    for path, _, nested_items in os.walk(project_path):
        # there is no duplicates in single json, because json is a dict

        _look_for_configs(nested_items, resources_meta, path, bundle_name)

    meta_for_validation = _resolve_aliases(resources_meta)
    # check if all dependencies were described
    common_validator = VALIDATOR_BY_TYPE_MAPPING[ALL_TYPES]
    for name, meta in meta_for_validation.items():
        common_validator(resource_meta=meta, all_meta=meta_for_validation)

        resource_type = meta['resource_type']
        type_validator = VALIDATOR_BY_TYPE_MAPPING.get(resource_type)
        if type_validator:
            type_validator(name, meta)

    return resources_meta


def _resolve_names_in_meta(resources_dict, old_value, new_value):
    if isinstance(resources_dict, dict):
        for k, v in resources_dict.items():
            if k in ['prefix', 'suffix']:
                continue
            if isinstance(v, str) and old_value == v:
                resources_dict[k] = v.replace(old_value, new_value)
            elif isinstance(v, str) and old_value in v and v.startswith('arn'):
                resources_dict[k] = v.replace(old_value, new_value)
            else:
                _resolve_names_in_meta(v, old_value, new_value)
    elif isinstance(resources_dict, list):
        for item in resources_dict:
            if isinstance(item, dict):
                _resolve_names_in_meta(item, old_value, new_value)
            elif isinstance(item, str):
                if item == old_value:
                    index = resources_dict.index(old_value)
                    del resources_dict[index]
                    resources_dict.append(new_value)


def create_meta(project_path, bundle_name):
    # create overall meta.json with all resource meta info
    meta_path = build_path(project_path, ARTIFACTS_FOLDER,
                           bundle_name)
    _LOG.info("Bundle path: {0}".format(meta_path))
    overall_meta = create_resource_json(project_path=project_path,
                                        bundle_name=bundle_name)
    bundle_dir = resolve_bundle_directory(bundle_name=bundle_name)
    write_content_to_file(bundle_dir, BUILD_META_FILE_NAME, overall_meta)


def resolve_meta(overall_meta):
    from syndicate.core import CONFIG
    iam_suffix = CONFIG.iam_suffix
    extended_prefix_mode = CONFIG.extended_prefix_mode
    overall_meta = _resolve_aliases(overall_meta)
    _LOG.debug('Resolved meta was created')
    _LOG.debug(prettify_json(overall_meta))
    _resolve_permissions_boundary(overall_meta)
    _LOG.debug('Permissions boundary were resolved')
    # get dict with resolved prefix and suffix in meta resources
    # key: current_name, value: resolved_name
    resolved_names = {}
    for name, res_meta in overall_meta.items():
        resource_type = res_meta['resource_type']
        if resource_type in GLOBAL_AWS_SERVICES or extended_prefix_mode:
            resolved_name = resolve_resource_name(
                resource_name=name,
                prefix=CONFIG.resources_prefix,
                suffix=CONFIG.resources_suffix)
            if resource_type == LAMBDA_TYPE:
                res_meta['prefix'] = CONFIG.resources_prefix
                res_meta['suffix'] = CONFIG.resources_suffix
            # add iam_suffix to IAM role only if it is specified in config file
            if resource_type == IAM_ROLE and iam_suffix:
                resolved_name = resolved_name + iam_suffix
            if name != resolved_name:
                resolved_names[name] = resolved_name
    _LOG.debug('Going to resolve names in meta')
    _LOG.debug('Resolved names mapping: {0}'.format(str(resolved_names)))
    for current_name, resolved_name in resolved_names.items():
        overall_meta[resolved_name] = overall_meta.pop(current_name)
        if not all([current_name, resolved_name]):
            continue
        _resolve_names_in_meta(overall_meta, current_name, resolved_name)
    return overall_meta


def _resolve_aliases(overall_meta):
    """
    :type overall_meta: dict
    """
    from syndicate.core import CONFIG
    if CONFIG.aliases:
        aliases = {'${' + key + '}': str(value) for key, value in
                   CONFIG.aliases.items()}
        overall_meta = resolve_dynamic_identifier(aliases, overall_meta)
    return overall_meta


def _resolve_permissions_boundary(overall_meta):
    """Adds to every resource with resource_type IAM_ROLE permissions boundary
    from the config"""
    from syndicate.core import CONFIG
    if CONFIG.iam_permissions_boundary:
        for name, meta in overall_meta.items():
            if meta.get('resource_type') == IAM_ROLE:
                meta['permissions_boundary'] = CONFIG.iam_permissions_boundary


def resolve_resource_name(resource_name, prefix=None, suffix=None):
    return _resolve_suffix_name(
        _resolve_prefix_name(resource_name, prefix), suffix)


def resolve_resource_name_by_data(resource_name, resource_prefix,
                                  resource_suffix):
    return _resolve_suffix_name(
        _resolve_prefix_name(resource_name, resource_prefix), resource_suffix)


def _resolve_prefix_name(resource_name, resource_prefix):
    if resource_prefix:
        return resolve_aliases_for_string(resource_prefix) + resource_name
    return resource_name


def _resolve_suffix_name(resource_name, resource_suffix):
    if resource_suffix:
        return resource_name + resolve_aliases_for_string(resource_suffix)
    return resource_name



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.project_state.project_state import PROJECT_STATE_FILE
from syndicate.core.project_state.project_state import ProjectState

_LOG = get_logger('sync-processor')
USER_LOG = get_user_logger()


def sync_project_state():
    from syndicate.core import CONFIG, CONN, PROJECT_STATE
    from pathlib import PurePath
    bucket_name = CONFIG.deploy_target_bucket
    key_compound = PurePath(CONFIG.deploy_target_bucket_key_compound,
                            PROJECT_STATE_FILE).as_posix()
    s3 = CONN.s3()
    if not s3.is_file_exists(bucket_name=bucket_name,
                             key=key_compound):
        _LOG.debug('Remote .syndicate file does not exists. Pushing...')
        PROJECT_STATE.save_to_remote()
        _LOG.debug('Push successful')

    else:
        _LOG.debug('Syncing with remote .syndicate file.')
        remote_project_state = ProjectState.get_remote()

        _LOG.debug('Actualizing the project state...')
        PROJECT_STATE.actualize_locks(remote_project_state)
        PROJECT_STATE.add_execution_events(remote_project_state.events)
        PROJECT_STATE.actualize_latest_deploy(remote_project_state)

        _LOG.debug('Saving a local .syndicate file.')
        PROJECT_STATE.save()

        _LOG.debug('Synced the .syndicate file. Pushing...')
        PROJECT_STATE.save_to_remote()
        _LOG.debug('Push successful')
        _LOG.debug('Remote .syndicate file pulled')
    _LOG.info('Project state file has been successfully synced')



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json

from syndicate.core.resources.cloud_watch_resource import \
    validate_cloud_watch_rule_params
from syndicate.core.transform.terraform.converter.tf_resource_converter import \
    TerraformResourceConverter
from syndicate.core.transform.terraform.tf_helper import deploy_regions

SNS_TARGET = 'sns'
SQS_TARGET = 'sqs'
LAMBDA_TARGET = 'lambda'


class CloudWatchRuleConverter(TerraformResourceConverter):

    def convert(self, name, resource):
        validate_cloud_watch_rule_params(name=name, meta=resource)

        rule_type = resource['rule_type']
        regions = deploy_regions(resource_name=name, meta=resource)
        func = RULE_TYPES[rule_type]
        default_region = self.config.region
        for region in regions:
            self.template.add_provider_if_not_exists(region=region)
            tf_resource_name = f'{name}_{region}'
            provider = None
            if region != default_region:
                provider_type = self.template.provider_type()
                provider = f'{provider_type}.{region}'

            func(template=self.template, rule_name=name,
                 resource=resource, resource_name=tf_resource_name,
                 provider=provider)


def event_pattern(event_source, event_names):
    resource = {
        "eventSource": [
            event_source
        ],
        "eventName": event_names
    }
    return resource


def _create_ec2_rule(template, rule_name, resource,
                     resource_name, provider=None):
    instances = resource.get('instances')
    instance_states = resource.get('instance_states')

    event_pattern = {
        "source": ["aws.ec2"],
        "detail-type": ["EC2 Instance State-change Notification"]
    }

    rule_meta = {
        "name": rule_name,
        "event_pattern": event_pattern
    }

    if provider:
        rule_meta['provider'] = provider

    if instances:
        event_pattern["detail"] = {"instance-id": instances}
    if instance_states:
        if event_pattern.get("detail"):
            event_pattern.get("detail").update({"state": instance_states})
        else:
            event_pattern["detail"] = {"state": instance_states}

    rule = {
        resource_name: rule_meta
    }
    template.add_aws_cloudwatch_event_rule(meta=rule)


def _create_schedule_rule(template, rule_name, resource, resource_name,
                          provider=None):
    expression = resource.get('expression')

    rule_meta = {
        "name": rule_name,
        "schedule_expression": expression
    }

    if provider:
        rule_meta['provider'] = provider

    rule = {
        resource_name: rule_meta
    }
    template.add_aws_cloudwatch_event_rule(meta=rule)


def _create_api_call_rule(template, rule_name, resource, resource_name,
                          provider=None):
    operations = resource.get('operations')
    aws_service = resource.get('aws_service')
    custom_pattern = resource.get('custom_pattern')
    if custom_pattern:
        pattern = custom_pattern
    elif aws_service:
        pattern = {
            'detail-type': [
                'AWS API Call via CloudTrail'
            ],
            'detail': {
                'eventSource': [
                    '{0}.amazonaws.com'.format(aws_service)
                ]
            }
        }
        if operations:
            pattern['detail']['eventName'] = operations

    rule_meta = {
        "name": rule_name,
        "event_pattern": json.dumps(pattern)
    }

    if provider:
        rule_meta['provider'] = provider

    rule = {
        resource_name: rule_meta
    }
    template.add_aws_cloudwatch_event_rule(meta=rule)


RULE_TYPES = {
    'schedule': _create_schedule_rule,
    'ec2': _create_ec2_rule,
    'api_call': _create_api_call_rule
}



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from troposphere import GetAtt, Ref, logs, awslambda

from syndicate.commons.log_helper import get_user_logger
from syndicate.connection.cloud_watch_connection import \
    get_lambda_log_group_name
from syndicate.core.constants import S3_PATH_NAME
from syndicate.core.resources.helper import validate_params
from syndicate.core.resources.lambda_resource import \
    (LAMBDA_CONCUR_QUALIFIER_VERSION, LambdaResource,
     LAMBDA_MAX_CONCURRENCY, PROVISIONED_CONCURRENCY,
     LAMBDA_CONCUR_QUALIFIER_ALIAS,
     DYNAMODB_TRIGGER_REQUIRED_PARAMS, SQS_TRIGGER_REQUIRED_PARAMS,
     CLOUD_WATCH_TRIGGER_REQUIRED_PARAMS, S3_TRIGGER_REQUIRED_PARAMS,
     SNS_TRIGGER_REQUIRED_PARAMS, KINESIS_TRIGGER_REQUIRED_PARAMS)
from syndicate.core.resources.s3_resource import S3Resource
from .cf_cloudwatch_rule_converter import attach_rule_target
from .cf_dynamodb_table_converter import CfDynamoDbTableConverter
from .cf_iam_role_converter import CfIamRoleConverter
from .cf_resource_converter import CfResourceConverter
from .cf_s3_converter import CfS3Converter
from .cf_sns_converter import CfSnsConverter
from ..cf_transform_utils import \
    (to_logic_name, lambda_publish_version_logic_name,
     lambda_alias_logic_name, lambda_function_logic_name,
     iam_role_logic_name, dynamodb_table_logic_name,
     sqs_queue_logic_name, cloudwatch_rule_logic_name,
     s3_bucket_logic_name, sns_topic_logic_name,
     kinesis_stream_logic_name, lambda_layer_logic_name)

_LOG = get_user_logger()


class CfLambdaFunctionConverter(CfResourceConverter):

    def convert(self, name, meta):
        lambda_function = awslambda.Function(lambda_function_logic_name(name))
        lambda_function.FunctionName = meta['name']
        lambda_function.Code = awslambda.Code(
            S3Bucket=self.config.deploy_target_bucket,
            S3Key=meta[S3_PATH_NAME])
        lambda_function.Handler = meta['func_name']
        lambda_function.MemorySize = meta['memory']
        role_name = meta['iam_role_name']

        role = self.get_resource(iam_role_logic_name(role_name))
        _arn = None
        if not role:
            _LOG.warning(f'Role \'{role_name}\' was not found in '
                         f'build_meta.json. Building arn manually')
            iam_service = self.resources_provider.iam()
            role = role_name  # for kinesis trigger
            _arn = iam_service.iam_conn.build_role_arn(role_name)
        else:
            _arn = GetAtt(iam_role_logic_name(meta['iam_role_name']), 'Arn')
        lambda_function.Role = _arn
        lambda_function.Runtime = meta['runtime'].lower()
        lambda_function.Timeout = meta['timeout']

        env_vars = meta.get('env_variables')
        if env_vars:
            lambda_function.Environment = \
                awslambda.Environment(Variables=env_vars)

        dl_target_arn = LambdaResource.get_dl_target_arn(
            meta=meta,
            region=self.config.region,
            account_id=self.config.account_id)
        if dl_target_arn:
            lambda_function.DeadLetterConfig = \
                awslambda.DeadLetterConfig(TargetArn=dl_target_arn)

        layer_meta = meta.get('layers')
        if layer_meta:
            lambda_layers = []
            for layer_name in layer_meta:
                layer_logic_name = lambda_layer_logic_name(layer_name)
                layer = self.get_resource(layer_logic_name)
                if not layer:
                    raise AssertionError("Lambda layer '{}' is not present "
                                         "in build meta.".format(layer_name))
                lambda_layers.append(layer.ref())
            lambda_function.Layers = lambda_layers

        vpc_sub_nets = meta.get('subnet_ids')
        vpc_security_group = meta.get('security_group_ids')
        if vpc_sub_nets and vpc_security_group:
            lambda_function.VpcConfig = awslambda.VPCConfig(
                SubnetIds=vpc_sub_nets,
                SecurityGroupIds=vpc_security_group)

        tracing_mode = meta.get('tracing_mode')
        if tracing_mode:
            lambda_function.TracingConfig = \
                awslambda.TracingConfig(Mode=tracing_mode)

        reserved_concur = meta.get(LAMBDA_MAX_CONCURRENCY)
        lambda_service = self.resources_provider.lambda_resource()
        if lambda_service.check_concurrency_availability(reserved_concur):
            lambda_function.ReservedConcurrentExecutions = reserved_concur

        self.template.add_resource(lambda_function)

        provisioned_concur = meta.get(PROVISIONED_CONCURRENCY)

        publish_version = meta.get('publish_version', False)
        lambda_version = None
        if publish_version:
            lambda_version = self._lambda_version(
                function=lambda_function,
                provisioned_concurrency=provisioned_concur)

        alias = meta.get('alias')
        lambda_alias = None
        if alias:
            lambda_alias = self._lambda_alias(
                function=lambda_function, alias=alias,
                version_logic_name=lambda_version.title,
                provisioned_concurrency=provisioned_concur)

        retention = meta.get('logs_expiration')
        if retention:
            group_name = get_lambda_log_group_name(lambda_name=name)
            self.template.add_resource(
                self._log_group(group_name=group_name,
                                retention_in_days=retention))
        event_sources = meta.get('event_sources')
        if event_sources:
            for trigger_meta in event_sources:
                if lambda_alias:
                    arn = lambda_alias.ref()
                elif lambda_version:
                    arn = lambda_version.ref()
                else:
                    arn = lambda_function.ref()

                trigger_type = trigger_meta['resource_type']
                func = self.CREATE_TRIGGER[trigger_type]
                func(self, name, arn, role, trigger_meta)

    def _create_dynamodb_trigger_from_meta(self, lambda_name, lambda_arn,
                                           role, trigger_meta):
        validate_params(lambda_name, trigger_meta,
                        DYNAMODB_TRIGGER_REQUIRED_PARAMS)
        table_name = trigger_meta['target_table']

        table = self.get_resource(dynamodb_table_logic_name(table_name))

        if not CfDynamoDbTableConverter.is_stream_enabled(table):
            CfDynamoDbTableConverter.configure_table_stream(table)
        self.template.add_resource(self._event_source_mapping(
            lambda_arn=lambda_arn,
            lambda_name=lambda_name,
            event_source_arn=table.get_att('StreamArn'),
            event_source_name=table_name,
            batch_size=trigger_meta['batch_size'],
            starting_position='LATEST'))

    def _create_sqs_trigger_from_meta(self, lambda_name, lambda_arn, role,
                                      trigger_meta):
        validate_params(lambda_name, trigger_meta, SQS_TRIGGER_REQUIRED_PARAMS)
        target_queue_name = trigger_meta['target_queue']

        queue_logic_name = sqs_queue_logic_name(target_queue_name)
        target_queue = self.get_resource(queue_logic_name)
        if not target_queue:
            _LOG.error('Queue {} does not exist'.format(target_queue))
            return
        self.template.add_resource(self._event_source_mapping(
            lambda_arn=lambda_arn,
            lambda_name=lambda_name,
            event_source_arn=target_queue.get_att('Arn'),
            event_source_name=target_queue_name,
            batch_size=trigger_meta['batch_size']))

    def _create_cloud_watch_trigger_from_meta(self, lambda_name, lambda_arn,
                                              role, trigger_meta):
        validate_params(lambda_name, trigger_meta,
                        CLOUD_WATCH_TRIGGER_REQUIRED_PARAMS)
        rule_name = trigger_meta['target_rule']
        rule = self.get_resource(cloudwatch_rule_logic_name(rule_name))

        attach_rule_target(rule=rule, target_arn=lambda_arn)
        permission = self.convert_lambda_permission(
            lambda_arn=lambda_arn,
            lambda_name=lambda_name,
            principal='events',
            source_arn=rule.get_att('Arn'),
            permission_qualifier=rule_name
        )
        self.template.add_resource(permission)

    def _create_s3_trigger_from_meta(self, lambda_name, lambda_arn, role,
                                     trigger_meta):
        validate_params(lambda_name, trigger_meta, S3_TRIGGER_REQUIRED_PARAMS)
        target_bucket_name = trigger_meta['target_bucket']

        bucket_logic_name = s3_bucket_logic_name(target_bucket_name)
        target_bucket = self.get_resource(bucket_logic_name)
        if not target_bucket:
            _LOG.error('S3 bucket {0} event source for lambda {1} '
                       'was not created.'.format(target_bucket, lambda_name))
            return
        permission = self.convert_lambda_permission(
            lambda_arn=lambda_arn,
            lambda_name=lambda_name,
            principal='s3',
            source_arn=S3Resource.get_bucket_arn(target_bucket_name),
            permission_qualifier=target_bucket_name
        )
        target_bucket.DependsOn = permission
        self.template.add_resource(permission)
        CfS3Converter.configure_event_source_for_lambda(
            bucket=target_bucket,
            lambda_arn=lambda_arn,
            events=trigger_meta['s3_events'],
            filter_rules=trigger_meta.get('filter_rules')
        )

    def _create_sns_topic_trigger_from_meta(self, lambda_name, lambda_arn,
                                            role, trigger_meta):
        validate_params(lambda_name, trigger_meta, SNS_TRIGGER_REQUIRED_PARAMS)
        topic_name = trigger_meta['target_topic']

        topic = self.get_resource(sns_topic_logic_name(topic_name))
        if not topic:
            raise AssertionError(
                'Topic does not exist: {0}.'.format(topic_name))

        # region = trigger_meta.get('region') TODO: support region param

        CfSnsConverter.subscribe(topic=topic,
                                 protocol='lambda',
                                 endpoint=lambda_arn)
        permission = self.convert_lambda_permission(
            lambda_arn=lambda_arn,
            lambda_name=lambda_name,
            principal='sns',
            source_arn=topic.ref(),
            permission_qualifier=topic_name
        )
        self.template.add_resource(permission)

    def _create_kinesis_stream_trigger_from_meta(self, lambda_name, lambda_arn,
                                                 role, trigger_meta):
        validate_params(lambda_name, trigger_meta,
                        KINESIS_TRIGGER_REQUIRED_PARAMS)

        stream_name = trigger_meta['target_stream']

        stream = self.get_resource(kinesis_stream_logic_name(stream_name))
        if not stream:
            _LOG.error('Kinesis stream does not exist: {0}.'
                       .format(stream_name))
            return

        policy_name = '{0}KinesisTo{1}Lambda'.format(stream_name, lambda_name)
        policy_document = {
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "lambda:InvokeFunction"
                    ],
                    "Resource": [
                        lambda_arn
                    ]
                },
                {
                    "Action": [
                        "kinesis:DescribeStreams",
                        "kinesis:DescribeStream",
                        "kinesis:ListStreams",
                        "kinesis:GetShardIterator",
                        "Kinesis:GetRecords"
                    ],
                    "Effect": "Allow",
                    "Resource": stream.get_att('Arn')
                }
            ],
            "Version": "2012-10-17"
        }
        policy = CfIamRoleConverter.convert_inline_policy(
            role=role,
            policy_name=policy_name,
            policy_document=policy_document)
        self.template.add_resource(policy)
        event_mapping = self._event_source_mapping(
            lambda_arn=lambda_arn,
            lambda_name=lambda_name,
            event_source_arn=stream.get_att('Arn'),
            event_source_name=stream_name,
            batch_size=trigger_meta['batch_size'],
            starting_position=trigger_meta['starting_position'])
        event_mapping.DependsOn = policy
        self.template.add_resource(event_mapping)

    @staticmethod
    def _event_source_mapping(lambda_arn, lambda_name,
                              event_source_arn, event_source_name,
                              batch_size, starting_position=None):
        event_source = awslambda.EventSourceMapping(to_logic_name(
            'LambdaEventSourceMapping', lambda_name, event_source_name))
        event_source.BatchSize = batch_size
        event_source.Enabled = True
        event_source.EventSourceArn = event_source_arn
        event_source.FunctionName = lambda_arn
        if starting_position:
            event_source.StartingPosition = starting_position
        return event_source

    CREATE_TRIGGER = {
        'dynamodb_trigger': _create_dynamodb_trigger_from_meta,
        'cloudwatch_rule_trigger': _create_cloud_watch_trigger_from_meta,
        'eventbridge_rule_trigger': _create_cloud_watch_trigger_from_meta,
        's3_trigger': _create_s3_trigger_from_meta,
        'sns_topic_trigger': _create_sns_topic_trigger_from_meta,
        'kinesis_trigger': _create_kinesis_stream_trigger_from_meta,
        'sqs_trigger': _create_sqs_trigger_from_meta
    }

    def _lambda_version(self, function, provisioned_concurrency=None):
        version_name = lambda_publish_version_logic_name(function.FunctionName)
        lambda_version = awslambda.Version(version_name)
        lambda_version.FunctionName = Ref(function)
        self._add_provisioned_concur(
            resource=lambda_version,
            provisioned_concurrency=provisioned_concurrency,
            expected_qualifier=LAMBDA_CONCUR_QUALIFIER_VERSION)
        self.template.add_resource(lambda_version)
        return lambda_version

    def _lambda_alias(self, function, alias, version_logic_name,
                      provisioned_concurrency=None):
        alias_resource_name = \
            lambda_alias_logic_name(function_name=function.FunctionName,
                                    alias=alias)
        lambda_alias = awslambda.Alias(alias_resource_name)
        lambda_alias.FunctionName = Ref(function)
        lambda_alias.Name = alias
        lambda_alias.FunctionVersion = GetAtt(version_logic_name, 'Version') \
            if version_logic_name else '$LATEST'

        self._add_provisioned_concur(
            resource=lambda_alias,
            provisioned_concurrency=provisioned_concurrency,
            expected_qualifier=LAMBDA_CONCUR_QUALIFIER_ALIAS)
        self.template.add_resource(lambda_alias)
        return lambda_alias

    @staticmethod
    def _add_provisioned_concur(resource, provisioned_concurrency,
                                expected_qualifier):
        if provisioned_concurrency:
            qualifier = provisioned_concurrency.get('qualifier')
            if qualifier == expected_qualifier:
                value = provisioned_concurrency.get('value')
                resource.ProvisionedConcurrencyConfig = \
                    awslambda.ProvisionedConcurrencyConfiguration(
                        ProvisionedConcurrentExecutions=value)

    @staticmethod
    def _log_group(group_name, retention_in_days):
        name = to_logic_name('LogsLogGroup', group_name)
        log_group = logs.LogGroup(name)
        log_group.LogGroupName = group_name
        log_group.RetentionInDays = retention_in_days
        return log_group

    @staticmethod
    def convert_lambda_permission(lambda_arn, lambda_name, principal,
                                  source_arn=None, permission_qualifier=''):
        lambda_permission = awslambda.Permission(to_logic_name(
            'LambdaPermission', lambda_name, principal, permission_qualifier))
        lambda_permission.FunctionName = lambda_arn
        lambda_permission.Action = 'lambda:InvokeFunction'
        lambda_permission.Principal = '{}.amazonaws.com'.format(principal)
        if source_arn:
            lambda_permission.SourceArn = source_arn
        return lambda_permission



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from concurrent.futures import ThreadPoolExecutor, as_completed

from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.connection import ResourceGroupsTaggingAPIConnection
from syndicate.core.constants import LAMBDA_TYPE, SWAGGER_UI_TYPE, \
    EC2_LAUNCH_TEMPLATE_TYPE
from syndicate.core.resources.helper import chunks

_LOG = get_logger('syndicate.core.resources.group_tagging_api_resource')
USER_LOG = get_user_logger()


class TagsApiResource:
    def __init__(self, connection: ResourceGroupsTaggingAPIConnection):
        self.connection = connection
        self.resource_type_to_preprocessor_mapping = {
            LAMBDA_TYPE: self._preprocess_lambda_arn
        }
        from syndicate.core import CONFIG
        self.tags = CONFIG.tags

    @staticmethod
    def _preprocess_lambda_arn(arn: str, meta: dict):
        """Extracts lambda's arn without version and alias """
        lambda_name = meta['resource_name']
        processed_arn = arn[:arn.index(lambda_name) + len(lambda_name)]
        _LOG.debug(f'Processing lambda\'s arn from \'{arn}\' '
                   f'to \'{processed_arn}\'')
        return processed_arn

    def _extract_arns(self, output: dict) -> list:
        arns = []
        _LOG.info(f'Extracting and processing arns from output')
        for arn, meta in output.items():
            resource_type = meta['resource_meta']['resource_type']
            if resource_type in [SWAGGER_UI_TYPE, EC2_LAUNCH_TEMPLATE_TYPE]:
                continue
            processor_func = self.resource_type_to_preprocessor_mapping.get(
                resource_type)
            if processor_func:
                arn = processor_func(arn, meta)
            arns.append(arn)
        return arns

    def apply_tags(self, output: dict):
        if not self.tags:
            USER_LOG.info('No tags are specified in config. Skipping...')
            return
        arns = self._extract_arns(output)
        failed_arns = []
        with ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self.connection.tag_resources, batch,
                                self.tags) for batch in chunks(arns, 20)
            ]
            for future in as_completed(futures):
                failed = future.result()
                if failed:
                    failed_arns.extend(failed.keys())
        if not failed_arns:
            USER_LOG.info(f'Tags: {self.tags} were successfully applied to '
                          f'all resources')
        else:
            USER_LOG.warn(f'Couldn\'t apply tags for resources: '
                          f'{failed_arns}')

    def remove_tags(self, output: dict):
        if not self.tags:
            USER_LOG.info('No tags are specified in config. Skipping...')
            return
        arns = self._extract_arns(output)
        failed_arns = []
        with ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(
                    self.connection.untag_resources, batch,
                    list(self.tags.keys())) for batch in chunks(arns, 20)
            ]
            for future in as_completed(futures):
                failed = future.result()
                if failed:
                    failed_arns.extend(failed.keys())
        if not failed_arns:
            USER_LOG.info(f'Tags {list(self.tags.keys())} were successfully '
                          f'removed from all resources')
        else:
            USER_LOG.warn(f'Couldn\'t remove tags from resources: '
                          f'{failed_arns}')



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""


def build_function_name_ref(function_name):
    return '${aws_lambda_function.' + function_name + '.function_name}'


def build_function_arn_ref(function_name):
    return '${aws_lambda_function.' + function_name + '.arn}'


def build_lambda_alias_name_ref(alias_name):
    return '${aws_lambda_alias.' + alias_name + '.name}'


def build_lambda_version_ref(function_name):
    return '${aws_lambda_function.' + function_name + '.version}'


def build_function_invoke_arn_ref(function_name):
    return '${aws_lambda_function.' + function_name + '.invoke_arn}'


def build_authorizer_id_ref(authorizer_name):
    return '${aws_api_gateway_authorizer.' + authorizer_name + '.id}'


def build_api_gateway_method_name_reference(method_name):
    return '${aws_api_gateway_method.' + method_name + '.http_method}'


def build_api_gateway_resource_id_ref(resource_name):
    return '${aws_api_gateway_resource.' + resource_name + '.id}'


def build_api_gateway_resource_path_ref(resource_name):
    return '${aws_api_gateway_resource.' + resource_name + '.path}'


def build_rest_api_id_ref(api_name):
    return '${aws_api_gateway_rest_api.' + api_name + '.id}'


def build_sns_topic_arn_ref(sns_topic):
    return '${aws_sns_topic.' + sns_topic + '.arn}'


def build_api_gateway_root_resource_id_ref(api_name):
    return '${aws_api_gateway_rest_api.' + api_name + '.root_resource_id}'


def build_api_gateway_deployment_id_ref(deployment_name):
    return '${aws_api_gateway_deployment.' + deployment_name + '.id}'


def build_cognito_identity_pool_id(pool_name):
    return '${aws_cognito_identity_pool.' + pool_name + '.id}'


def build_ref_to_lambda_layer_arn(layer_name):
    return '${aws_lambda_layer_version.' + layer_name + '.arn}'


def build_role_arn_ref(role_name):
    return '${aws_iam_role.' + role_name + '.arn}'


def build_cloud_watch_event_rule_name_ref(target_rule):
    return '${aws_cloudwatch_event_rule.' + target_rule + '.name}'


def build_resource_arn_ref(tf_resource_type, name):
    return '${' + tf_resource_type + '.' + name + '.arn}'


def build_policy_arn_ref(policy_name):
    return '${aws_iam_policy.' + policy_name + '.arn}'


def build_com_env_arn(com_env_name):
    return '${aws_batch_compute_environment.' + com_env_name + '.arn}'


def build_role_name_ref(role_name):
    return '${aws_iam_role.' + role_name + '.name}'


def build_role_id_ref(role_name):
    return '${aws_iam_role.' + role_name + '.id}'


def build_instance_profile_arn_ref(instance_profile_name):
    return '${aws_iam_instance_profile.' + instance_profile_name + '.arn}'


def build_kinesis_stream_arn_ref(stream_name):
    return '${aws_kinesis_stream.' + stream_name + '.arn}'


def build_bucket_id_ref(bucket_name):
    return '${aws_s3_bucket.' + bucket_name + '.id}'


def build_bucket_arn_ref(bucket_name):
    return '${aws_s3_bucket.' + bucket_name + '.arn}'


def build_dynamo_db_stream_arn_ref(table_name):
    return '${aws_dynamodb_table.' + table_name + '.stream_arn}'


def build_sqs_queue_arn_ref(queue_name):
    return '${aws_sqs_queue.' + queue_name + '.arn}'


def build_sqs_queue_arn_ref(queue_name):
    return '${aws_sqs_queue.' + queue_name + '.arn}'


def build_sqs_queue_id_ref(queue_name):
    return '${aws_sqs_queue.' + queue_name + '.id}'


def build_aws_appautoscaling_target_resource_id_ref(target_name):
    return '${aws_appautoscaling_target.' + target_name + '.resource_id}'


def build_aws_appautoscaling_target_scalable_dimension_ref(target_name):
    return '${aws_appautoscaling_target.' + target_name + '.scalable_dimension}'


def build_aws_appautoscaling_target_service_namespace_ref(target_name):
    return '${aws_appautoscaling_target.' + target_name + '.service_namespace}'



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
import shutil

from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import build_path, execute_command_by_path

_LOG = get_logger('java_runtime_assembler')

MVN_TARGET_DIRECTORY = 'target'


def assemble_java_mvn_lambdas(project_path, bundles_dir):
    from syndicate.core import CONFIG
    src_path = build_path(CONFIG.project_path, project_path)
    _LOG.info(f'Java sources are located by path: {src_path}')
    _LOG.info(f'Going to process java mvn project by path: '
              f'{CONFIG.project_path}')
    execute_command_by_path(command='mvn clean install',
                            path=CONFIG.project_path)

    # copy java artifacts to the target folder
    for root, dirs, files in os.walk(os.path.join(CONFIG.project_path,
                                                  MVN_TARGET_DIRECTORY)):
        for file in files:
            if file.endswith(".jar") or file.endswith(".war") \
                    or file.endswith(".zip"):
                shutil.copyfile(build_path(root, file),
                                build_path(bundles_dir, file))
    _LOG.info('Java mvn project was processed successfully')



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from botocore.exceptions import WaiterError

from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger('syndicate.core.resources.batch_jobqueue')

DEFAULT_STATE = 'ENABLED'


class BatchJobQueueResource(BaseResource):
    def __init__(self, batch_conn):
        self.batch_conn = batch_conn

    def create_job_queue(self, args):
        return self.create_pool(self._create_job_queue_from_meta, args)

    def describe_job_queue(self, name, meta):
        response = self.batch_conn.describe_job_queue(name)
        try:
            arn = response['jobQueues'][0]['jobQueueArn']
            return {arn: build_description_obj(response, name, meta)}
        except (KeyError, IndexError):
            _LOG.warn("Batch Job Queue %s not found", name)
            return {}

    def remove_job_queue(self, args):
        self.create_pool(self._remove_job_queue, args)

    @unpack_kwargs
    def _remove_job_queue(self, arn, config):
        job_queue_name = config.get('resource_name')

        self.batch_conn.update_job_queue(
            job_queue=arn,
            state='DISABLED',
            compute_environment_order=[]
        )
        self.batch_conn.delete_job_queue(job_queue=arn)
        _LOG.info('Batch Job Queue %s was removed.', job_queue_name)

    @unpack_kwargs
    def _create_job_queue_from_meta(self, name, meta):
        params = meta.copy()
        params['job_queue_name'] = name
        if 'resource_type' in params:
            del params['resource_type']

        if self._is_job_queue_exist(name):
            _LOG.warn('Batch Job Queue %s already exists', name)
            return self.describe_job_queue(name, meta)

        state = params.get('state')
        if not state:
            params['state'] = DEFAULT_STATE

        self.batch_conn.create_job_queue(**params)
        try:
            waiter = self.batch_conn.get_job_queue_waiter()
            waiter.wait(jobQueues=[name])
        except WaiterError as e:
            _LOG.error(e)

        _LOG.info('Created Batch Job Queue %s.', name)
        return self.describe_job_queue(name, meta)

    def _is_job_queue_exist(self, job_queue_name):
        response = self.batch_conn.describe_job_queue(
            job_queues=job_queue_name
        )
        return bool(response['jobQueues'])
